#!/usr/bin/env python

import os
import sys

def stripList(l):
   return([x.strip() for x in l])

def executeCommand(command):
   output = os.popen(command, 'rb').readlines()
   return stripList(output)

def brokenLink(link):
   if not os.path.exists(link):
      os.unlink(link)
      if not os.listdir(os.path.dirname(link)):
         os.removedirs(os.path.dirname(link))
      return True
   else:
      return False

def renameSyms(shorten, inModules, rootPath):
   # Rename all symbols in the shorten dictionary for all inModules
   kalSrcDir = open('/etc/natinst/nikal/nikal.dir', 'r').read().strip() + '/src'
   fh = open('%s/shortenNISyms.txt' % kalSrcDir, 'w')
   # For each module in the shorten dictionary find the symbols to rename
   for file in shorten:
      modname = os.path.basename(file).rstrip('-unversioned.o')
      # For each symbol in the list create an enumerated shortend name
      # and write it to file for objcopy to parse later
      for i, symbol in enumerate(shorten[file]):
         fh.write('%s %s\n' % (symbol, modname + '_exported' + str(i)))
   fh.close()
   outModules = []
   # Use objcopy to rename all symbols in the shortenNISyms.txt file
   for i, module in enumerate(inModules):
      #modname = os.path.basename(module)
      modname = module.replace(rootPath + '/', '')
      outModules.append('/tmp/strippedNIModules/' + modname)
      if not os.path.isdir(os.path.dirname(outModules[i])):
         os.makedirs(os.path.dirname(outModules[i]))
      executeCommand('objcopy --redefine-syms=%s/shortenNISyms.txt %s %s' % (kalSrcDir, module, outModules[i]))
   return outModules

def sortModules(modules, undefined, exported, shorten):
   # Sort the input modules list in dependency order
   done = False
   while not done:
      done = True
      for module in undefined:
         currentIndex = modules.index(module)
         maxIndex = 0
         for symbol in undefined[module]:
            if symbol in exported:
               # We can't have symbols longer than 56 characters so add them to
               # a dicionary to shorten later
               if len(symbol) > 56 and symbol not in shorten.get(exported[symbol], []):
                  shorten.setdefault(exported[symbol], []).append(symbol)
               if modules.index(exported[symbol]) > maxIndex:
                  maxIndex = modules.index(exported[symbol])
         # Found the exported symbol reorder the modules list if needed
         if maxIndex > currentIndex:
            modules.insert(maxIndex + 1, module)
            modules.pop(currentIndex)
            # A reorder was required so we may need another pass
            done = False

if __name__ == "__main__":
   if len(sys.argv) != 2:
      sys.exit('%s <nikalClientDB>' % sys.argv[0])

   rootPath = sys.argv[1]
   modules = executeCommand('find %s -type l' % rootPath)
   modules = [l for l in modules if not brokenLink(l)]

   shorten = {}
   undefined = {}
   exported = {}
   for module in modules:
      undefined[module] = executeCommand("nm -u %s | awk '{print $2}'" % module)
      exportedSyms = executeCommand("strings %s" % module)
      exportedSyms = exportedSyms[exportedSyms.index('nNIExport_beginExportedSymbolsArray') + 1:exportedSyms.index('nNIExport_endExportedSymbolsArray')]
      exported.update(dict(zip(exportedSyms, [module] * len(exportedSyms))))

   sortModules(modules, undefined, exported, shorten)
   modulesWithShortSyms = renameSyms(shorten, modules, rootPath)

   for module in modulesWithShortSyms:
      print module
