/**  This file implements the NI-KAL Kernel Abstraction Layer for Linux

    Copyright 2002-2005,
    National Instruments Corporation.
    All rights reserved.

    originated:  17.jun.2002
**/




#undef __NO_VERSION__

#define ___nikal_c___

#include <linux/version.h>

#define nNIKAL100_mIsLinux24 (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0) && (LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)))
#define nNIKAL100_mIsLinux26 (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) && (LINUX_VERSION_CODE < KERNEL_VERSION(2,7,0)))

#define nNIKAL180_mUSBSupportedVersion (nNIKAL100_mIsLinux26 && (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25)))
#define nNIKAL100_mUSBIsSupported (nNIKAL180_mUSBSupportedVersion && (defined(CONFIG_USB) || defined(CONFIG_USB_MODULE)))

#define nNIKAL100_mMin(x,y) ((x)<(y)?(x):(y))
#define nNIKAL100_mMax(x,y) ((x)>(y)?(x):(y))

#if !nNIKAL100_mIsLinux24 && !nNIKAL100_mIsLinux26
   #error ("NI-KAL currently only supports Linux kernel 2.4 and Linux kernel 2.6");
#endif

#ifdef nNIKAL160_kConfig
   #include <linux/config.h>
#endif
#ifdef MODULE
   #if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
      #define MODVERSIONS
   #endif

   #ifdef MODVERSIONS
      #if nNIKAL100_mIsLinux24
         #include <linux/modversions.h>
      #else
         #include <config/modversions.h>
      #endif
   #endif

   #include <linux/module.h>
   #include <linux/autoconf.h>
   #include <linux/kmod.h>
#endif

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#include <linux/list.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/usb.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/timer.h>
#include <linux/pagemap.h>
#ifdef nNIKAL150_kKthread
   #include <linux/kthread.h>
#endif
#ifdef nNIKAL150_kMutexMethod
   #include <linux/mutex.h>
#endif

#include <asm/mman.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/atomic.h>

#if nNIKAL100_mIsLinux24
   #include <linux/iobuf.h>
   #include <asm/softirq.h>
   #include <linux/devfs_fs_kernel.h>
#else
   #include <linux/err.h>
   #include <linux/workqueue.h>
   #include <linux/cpumask.h>
   #include <linux/pagemap.h>
   #include <linux/interrupt.h>
   #include <linux/cdev.h>
   #include <linux/gfp.h>
   #include <linux/jiffies.h>
#ifdef nNIKAL170_kIoctl32
   #include <linux/ioctl32.h>
#endif
#endif

#ifndef nNIKAL100_kPutPageTestzero
#define put_page_testzero(p)   atomic_dec_and_test(&(p)->count)
#endif

#include "nikal.h"

#ifdef MODULE_LICENSE
   MODULE_LICENSE("Copyright (c) 2002-2009 National Instruments Corporation.  All Rights Reserved.  Any and all use of the copyrighted materials is subject to the then current terms and conditions of the applicable license agreement, which can be found at <http://www.ni.com/linux/>.");
#endif

#ifdef MODULE_AUTHOR
   MODULE_AUTHOR("National Instruments Corporation");
#endif

#ifdef MODULE_DESCRIPTION
   MODULE_DESCRIPTION("National Instruments Kernel Abstraction Layer");
#endif

#if nNIKAL100_mIsLinux24
   #define nNIKAL100_MOD_INC_USE_COUNT MOD_INC_USE_COUNT
   #define nNIKAL100_MOD_DEC_USE_COUNT MOD_DEC_USE_COUNT

   #define nNIKAL100_tWorkQueue struct tq_struct
   #define nNIKAL100_mInitWorkQueue INIT_TQUEUE
   #define nNIKAL100_mScheduleWorkQueue schedule_task
#else
   #define nNIKAL100_MOD_INC_USE_COUNT
   #define nNIKAL100_MOD_DEC_USE_COUNT

   #define nNIKAL100_tWorkQueue struct work_struct
   #define nNIKAL100_mInitWorkQueue INIT_WORK
   #define nNIKAL100_mScheduleWorkQueue schedule_work
#endif


#define nNIKAL100_kPCIDeviceIDTableFooterSignature 0x70636964 
#define nNIKAL100_kUSBDeviceIDTableFooterSignature 0x75736274 
#ifdef KMALLOC_MAX_SIZE
#define nNIKAL100_kKMallocMaxSize KMALLOC_MAX_SIZE
#else
#define nNIKAL100_kKMallocMaxSize 0x20000 
#endif

#ifdef nNIKAL100_kUSBmsecTimeout
   #define nNIKAL100_kUSBControlTimeout USB_CTRL_GET_TIMEOUT
#else
   #define nNIKAL100_kUSBControlTimeout USB_CTRL_GET_TIMEOUT * HZ
#endif

#ifdef nNIKAL100_kDebuggingIsActive
#define nNIKAL100_kDebugPackageString "[nikal] "
#define nNIKAL100_assertFatal(test,msg)              \
           do                                        \
           {   if (!(test))                          \
               {                                     \
                  printk("[nikal] ASSERT: %s", msg); \
                  *(int*)NULL = 0;                   \
               }                                     \
           } while (0)
#define nNIKAL100_debugPrintf   printk("%s", nNIKAL100_kDebugPackageString),printk
#define nNIKAL120_mCheckStackUsage nNIKAL120_checkStackUsage()
#else
#define nNIKAL100_assertFatal(test,msg)
static inline void nNIKAL100___printfDoNothingHelper(const nNIKAL100_tText*fmt, ...) { }
#define nNIKAL100_debugPrintf   if(1) ; else nNIKAL100___printfDoNothingHelper
#define nNIKAL120_mCheckStackUsage
#endif


#ifdef URB_ASYNC_UNLINK
   #define nNIKAL130_kURBAsyncUnlink
#endif


#define nNIKAL120_mRoundUp(x, r) ((x + (r-1))&(~(r-1)))

/** Compile time assert from Jon Jagger
 * http://www.jaggersoft.com/pubs/CVu11_3.html **/
#define nNIKAL100_compileTimeAssert(pred,comment) switch(0){case 0:case pred:;}


#ifndef PCI_CAP_ID_SSVID
  #define PCI_CAP_ID_SSVID 0x0D
#endif

#define nNIKAL100_kSubsystemIDOffset       4
#define nNIKAL100_kSubsystemVendorIDMask   0x0000FFFF
#define nNIKAL100_kSubsystemDeviceIDShift  0x10





typedef struct sysinfo              nLinux_sysinfo;
typedef struct inode                nLinux_inode;
typedef struct file                 nLinux_fileHandle;
typedef struct vm_area_struct       nLinux_vmArea;
#ifdef nNIKAL190_kVmFaultMethod
typedef struct vm_fault             nLinux_vmFault;
#endif
typedef struct page                 nLinux_physicalPage;
typedef struct pt_regs              nLinux_registers;
typedef struct semaphore            nLinux_semaphore;
typedef struct tasklet_struct       nLinux_tasklet;
typedef struct pci_device_id        nLinux_pciDeviceID;
typedef struct usb_device_id        nLinux_usbDeviceID;
typedef struct mm_struct            nLinux_mm;
typedef struct timeval              nLinux_timeValue;
typedef struct timer_list           nLinux_timerList;
typedef struct pci_dev              nLinux_pciDevice;
typedef struct pci_bus              nLinux_pciBus;
typedef struct pci_driver           nLinux_pciDriver;
typedef struct usb_device           nLinux_usbDevice;
typedef struct usb_interface        nLinux_usbInterface;
typedef struct usb_driver           nLinux_usbDriver;
typedef struct usb_ctrlrequest      nLinux_usbCtrlRequest;
typedef struct urb                  nLinux_urb;
typedef struct usb_host_config      nLinux_usbConfigDescriptorWrapper;
typedef struct usb_host_interface   nLinux_usbInterfaceDescriptorWrapper;
typedef struct usb_host_endpoint    nLinux_usbEndpointDescriptorWrapper;
typedef struct usb_endpoint_descriptor nLinux_usbEndpointDescriptor;
typedef struct list_head            nLinux_listHead;
typedef struct module               nLinux_driver;
typedef struct file_operations      nLinux_fileOperations;
typedef struct vm_operations_struct nLinux_vmOperations;
typedef struct mutex                nLinux_mutex;
typedef struct task_struct          nLinux_task;
#if nNIKAL100_mIsLinux24
typedef struct kiobuf               nLinux_kiobuf;
#else
typedef struct cdev                 nLinux_cdev;
#endif
typedef        spinlock_t           nLinux_spinlock;





static int __init nNIKAL100_initDriver(void);
static void __exit nNIKAL100_cleanupDriver(void);
static int nNIKAL100_open(nLinux_inode *the_inode, nLinux_fileHandle *filePtr);
static int nNIKAL100_release(nLinux_inode *the_inode, nLinux_fileHandle *filePtr);
static int nNIKAL100_ioctl(nLinux_inode *the_inode, nLinux_fileHandle *filePtr, unsigned int command, unsigned long param);
#ifdef HAVE_COMPAT_IOCTL
static long nNIKAL100_compatIoctl(nLinux_fileHandle *filePtr, unsigned int command, unsigned long param);
#endif
static ssize_t nNIKAL100_write(nLinux_fileHandle *filePtr, const char __user *buffer, size_t count, loff_t *offset);
static ssize_t nNIKAL100_read(nLinux_fileHandle *filePtr, char __user *buffer, size_t count, loff_t *offset);
static int nNIKAL100_mmap(nLinux_fileHandle *filePtr, nLinux_vmArea *vma);
static loff_t nNIKAL100_llseek(nLinux_fileHandle *filePtr, loff_t inOffset, int seekFrom);

static void nNIKAL100_vmaOpen(nLinux_vmArea *vma);
static void nNIKAL100_vmaClose(nLinux_vmArea *vma);

static void nNIKAL100_dispatchDPC (unsigned long interruptData);

#if nNIKAL100_mIsLinux24
   static nLinux_physicalPage *nNIKAL100_vmaPageFaultHandler(nLinux_vmArea *vma,
      unsigned long address, int write_access);
   static void nNIKAL100_dispatchHardwareInterrupt (int irq, void *context,
      nLinux_registers *registers);
#else
#ifdef nNIKAL190_kVmFaultMethod
   static int nNIKAL190_vmaPageFaultHandler(nLinux_vmArea *vma, nLinux_vmFault *vmf);
#else
   static nLinux_physicalPage *nNIKAL100_vmaPageFaultHandler(nLinux_vmArea *vma,
      unsigned long address, int *type);
#endif
#ifdef nNIKAL160_kIRQRegs
   static irqreturn_t nNIKAL100_dispatchHardwareInterrupt (int irq, void *context,
      nLinux_registers *registers);
#else
   static irqreturn_t nNIKAL100_dispatchHardwareInterrupt (int irq, void *context);
#endif
#endif

static inline void nNIKAL100_listInit(nNIKAL100_tListNode *node)
{ node->prev = node->next = node; }

static inline void nNIKAL100_listAdd(nNIKAL100_tListNode *head, nNIKAL100_tListNode *node)
{
   node->prev = head->prev;
   node->next = head;
   head->prev->next = node;
   head->prev = node;
}

static inline void nNIKAL100_listRemove(nNIKAL100_tListNode *head, nNIKAL100_tListNode *node)
{
   node->prev->next = node->next;
   node->next->prev = node->prev;
   nNIKAL100_listInit(node);
}

static inline nNIKAL100_tBoolean nNIKAL100_listIsEmpty(nNIKAL100_tListNode *head)
{ return (head == head->next); }

static void *nNIKAL100_incrementPageRefcount(const void *ptr);
static void nNIKAL100_decrementPageRefcount(const void *ptr);
static nNIKAL100_tStatus nNIKAL100_remapPageRange(void *vmaPtr, nNIKAL100_tUPtr physAddr);




typedef struct
{
   pid_t callingThreadID;
   nNIKAL100_tMemoryAddressType addressType;
} nNIKAL100_tMMapData;

typedef struct
{
   nNIKAL100_tListNode head;
   nLinux_semaphore lock;
} nNIKAL100_tDriverList;

typedef struct
{
   nNIKAL100_tDriver *driver;
   void *deviceExtension;
} nNIKAL100_tDevicePrivateData;

typedef struct
{
   nNIKAL100_tDriver *driver;
   void *context;
} nNIKAL100_tFilePrivateData;

typedef nNIKAL100_tI32 (*nNIKAL100_cc nNIKAL100_tThreadFunction)(void *data);

typedef struct
{
   pid_t childThreadID;
   nNIKAL100_tThreadFunction threadFunction;
   void *threadContext;
   void * volatile threadCreatedEvent;
   atomic_t counter;
   nNIKAL100_tWorkQueue workQueue;
} nNIKAL100_tCreateThreadParamBlock;

#ifdef nNIKAL160_kWorkqueueNonDelay
typedef struct
{
   nNIKAL100_tWorkQueue workQueue;
   void *data;
} nNIKAL160_tWorkItemWithDataPointer;
#endif

typedef struct
{
   wait_queue_head_t waitQueue;
   nNIKAL100_tU32 done;
} nNIKAL100_tSingleUseEvent;

typedef struct
{
   nNIKAL100_tDPCCallback func;
   void *context;
} nNIKAL100_tDPCCallbackSpec;

typedef struct
{
   nLinux_tasklet tasklet;
   nNIKAL100_tDPCCallbackSpec callback;
} nNIKAL100_tDPC;

typedef struct
{
   nNIKAL100_tUPtr signature;
   nNIKAL100_tU32 capacity;
   nNIKAL100_tU32 size;
   nNIKAL100_tDriver *driver;
   nLinux_pciDeviceID *table;
} nNIKAL100_tPCIDeviceIDTableFooter;

#if nNIKAL100_mIsLinux26
typedef struct
{
   nNIKAL100_tUPtr signature;
   nNIKAL100_tU32 capacity;
   nNIKAL100_tU32 size;
   nNIKAL100_tDriver *driver;
   nLinux_usbDeviceID *table;
} nNIKAL100_tUSBDeviceIDTableFooter;
#endif

typedef struct
{
   nNIKAL100_tUPtr key;
   nNIKAL100_tUPtr value;
} nNIKAL100_tMapKeyValuePair;

#define nNIKAL100_tMapBufferNodeSize (128)

#define nNIKAL100_tMapBufferCapacity \
   ((nNIKAL100_tMapBufferNodeSize - sizeof(nNIKAL100_tListNode) - sizeof(nNIKAL100_tUPtr))/sizeof(nNIKAL100_tMapKeyValuePair))

typedef struct
{
   nNIKAL100_tListNode listNode;
   nNIKAL100_tUPtr size;
   nNIKAL100_tMapKeyValuePair keyValuePairs[nNIKAL100_tMapBufferCapacity];
} nNIKAL100_tMapBufferNode;

#define nNIKAL100_tMapHashSizeInBits 7
#define nNIKAL100_tMapHashSize  (1 << nNIKAL100_tMapHashSizeInBits)

typedef nNIKAL100_tListNode nNIKAL100_tMap[nNIKAL100_tMapHashSize];

typedef struct
{
   nNIKAL100_tMap *mapPtr;
   nNIKAL100_tListNode *bucketPtr;
   nNIKAL100_tMapBufferNode *bufferPtr; 
   nNIKAL100_tMapKeyValuePair *keyValuePtr;
} nNIKAL100_tMapIterator;

typedef struct
{
   nLinux_vmArea *vma;
   nNIKAL100_tUPtr address;
} nNIKAL100_tVMAAddressPair;

typedef void (*tFuncWithWorkQueuePtrParam)(nNIKAL100_tWorkQueue*);




static nLinux_semaphore nNIKAL100_sMMapSema;
static nLinux_semaphore nNIKAL100_sPageLockOperationSema;
#if nNIKAL100_mIsLinux24
static nNIKAL100_tU32 nNIKAL110_sJiffiesLastRead;
static nLinux_spinlock nNIKAL110_sJiffiesRolloverSpinlock;
#endif

static nNIKAL100_tMap nNIKAL100_sPageLockMap;
static nNIKAL100_tMap nNIKAL100_sMallocContiguousMap;
static nLinux_semaphore nNIKAL100_sMallocContiguousMapLock;

static const nNIKAL100_tMMapData nNIKAL100_gMMapData =
{
   .callingThreadID = 0,
   .addressType = nNIKAL100_kMemoryAddressTypeInvalid
};

static const nNIKAL100_tMMapData* nNIKAL100_gMMapDataPtr = &nNIKAL100_gMMapData;

static nNIKAL100_tDriverList nNIKAL100_driverList;

static nLinux_fileOperations nNIKAL100_fops =
{
   .open = nNIKAL100_open,
   .release = nNIKAL100_release,
   .ioctl = nNIKAL100_ioctl,
#ifdef HAVE_COMPAT_IOCTL
   .compat_ioctl = nNIKAL100_compatIoctl,
#endif
   .read = nNIKAL100_read,
   .write = nNIKAL100_write,
   .mmap = nNIKAL100_mmap,
   .llseek = nNIKAL100_llseek,
};

static nLinux_vmOperations nNIKAL100_vmaOps =
{
   .open = nNIKAL100_vmaOpen,
   .close = nNIKAL100_vmaClose,
#ifdef nNIKAL190_kVmFaultMethod
   .fault = nNIKAL190_vmaPageFaultHandler,
#else
   .nopage = nNIKAL100_vmaPageFaultHandler,
#endif
};

static const char *nNIVersion_versionStringArray[] =
{
   "nNIVersion_CompanyName=National Instruments Corporation",

   "nNIVersion_LegalCopyright=Copyright (c) 2002-2009 National Instruments Corporation. "
   "All Rights Reserved. Any and all use of the copyrighted materials is subject to the "
   "then current terms and conditions of the applicable license agreement, "
   "which can be found at <http://www.ni.com/linux/>.",

   "nNIVersion_ProductName=NIKAL",
   "nNIVersion_OriginalFilename=" nNIKAL100_kTargetName,
   "nNIVersion_FileDescription=NI-KAL Driver",

   #ifdef nNIKAL100_kDebuggingIsActive
      "nNIVersion_ProductVersion=1.10.0f0 debug build",
      "nNIVersion_FileVersion=1.10.0f0 debug build",
      "nNIVersion_InternalName=NIKAL 1.10.0f0 debug build",
   #else
      "nNIVersion_ProductVersion=1.10.0f0",
      "nNIVersion_FileVersion=1.10.0f0",
      "nNIVersion_InternalName=NIKAL 1.10.0f0",
   #endif

   NULL
};

#if nNIKAL100_mIsLinux24
   static devfs_handle_t sNipalkDevFSHandle;
#else
   static nLinux_cdev sNipalkCDev;
   static dev_t sNipalkDeviceNumber;
#endif

static nNIKAL100_tUPtr nNIKAL120_sStackSize;
static nNIKAL100_tUPtr nNIKAL120_sMaxStackUsageSeen = 0;
#define nNIKAL120_kMinStackUsageForWarning 0xc00  




static int nNIKAL100_gDeviceMajor;







inline nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_statusSelect(nNIKAL100_tStatus oldStatus, nNIKAL100_tStatus newStatus)
{
   if (nNIKAL100_statusIsFatal(oldStatus))
      return oldStatus;
   if (nNIKAL100_statusIsFatal(newStatus))
      return newStatus;
   if (oldStatus != 0)
      return oldStatus;
   return newStatus;
}

inline nNIKAL100_tStatus nNIKAL100_convertLinuxToKALStatus(int linuxStatus)
{
   switch(linuxStatus)
   {
      case 0:
         return nNIKAL100_kStatusSuccess;
         break;
      case -ENOMEM:
         return nNIKAL100_kStatusMemoryFull;
         break;
      case -ENAMETOOLONG:
         return nNIKAL100_kStatusValueConflict;
         break;
      case -EINVAL:
         return nNIKAL100_kStatusValueConflict;
         break;
      case -EINTR:
         return nNIKAL100_kStatusWaitInterrupted;
         break;
      case -EBUSY:
         return nNIKAL100_kStatusResourceBusy;
         break;
      case -EPERM:
         return nNIKAL100_kStatusOSFault;
         break;
      case -ECONNRESET:
         return nNIKAL100_kStatusTransferAborted;
         break;
      case -ENOENT:
         return nNIKAL100_kStatusTransferAborted;
         break;
      case -ENODEV:
         return nNIKAL100_kStatusDeviceNotFound;
         break;
      case -EPIPE:
         return nNIKAL100_kStatusCommunicationsFault;
         break;
      case -EMSGSIZE:
         return nNIKAL100_kStatusBadCount;
         break;
      case -EXDEV:
         return nNIKAL100_kStatusTransferStopped;
         break;
      case -ETIMEDOUT:
         return nNIKAL100_kStatusTransferTimedOut;
         break;
      case -ECOMM:
         return nNIKAL100_kStatusPhysicalBufferFull;
         break;
      case -ENOSR:
         return nNIKAL100_kStatusPhysicalBufferEmpty;
         break;
      case -EINPROGRESS:
         return nNIKAL100_kStatusTransferInProgress;
         break;
      case -EREMOTEIO:
         return nNIKAL100_kStatusCommunicationsFault;
         break;
      case -ENXIO:
         return nNIKAL100_kStatusBadMode;
         break;
      case -EFBIG:
         return nNIKAL100_kStatusBadCount;
         break;
      case -EAGAIN:
         return nNIKAL100_kStatusBadOffset;
         break;
      case -ENOSPC:
         return nNIKAL100_kStatusResourceBusy;
         break;
      
      case -ESHUTDOWN:
      case -EPROTO:
      case -EILSEQ:
      case -EOVERFLOW:
         return nNIKAL100_kStatusHardwareFault;
         break;
      case 256:
         
         return nNIKAL100_kStatusOSFault;
         break;
      default:
         nNIKAL100_debugPrintf("Unknown linux error code value!: %d\n", linuxStatus);
         nNIKAL100_assertFatal(0, "Unknown linux error code value!\n");
         return nNIKAL100_kStatusOSFault;
         break;
   }
}


static inline void nNIKAL100_mapInit(nNIKAL100_tMap *map)
{
   nNIKAL100_tU32 i;
   for (i=0; i< nNIKAL100_tMapHashSize; ++i)
   {
      nNIKAL100_listInit(&((*map)[i]));
   }
}


static void nNIKAL100_mapCleanup(nNIKAL100_tMap *map)
{
   nNIKAL100_tListNode *iterator, *temp;
   nNIKAL100_tMapBufferNode *bufferPtr;
   nNIKAL100_tU32 i, j;

   for (i=0; i< nNIKAL100_tMapHashSize; ++i)
   {
      if (nNIKAL100_listIsEmpty(&((*map)[i])))
         continue;

      iterator = (*map)[i].next;
      while (iterator != &((*map)[i]))
      {
         bufferPtr = (nNIKAL100_tMapBufferNode*)iterator;
         if (bufferPtr->size != 0)
         {
            for (j=0; j<bufferPtr->size; ++j)
            {
               nNIKAL100_debugPrintf("nNIKAL100_mapCleanup: found leaked map node: key: %lx, value: %lx, mapBuffer: %p\n", 
                  bufferPtr->keyValuePairs[j].key, bufferPtr->keyValuePairs[j].value, bufferPtr);
            }
         }
         
         temp = iterator;
         iterator = iterator->next;
         nNIKAL100_listRemove(&((*map)[i]), temp);

         nNIKAL100_debugPrintf("nNIKAL100_mapCleanup: deleting mapBuffer: %p\n", bufferPtr);
         nNIKAL100_free(bufferPtr);
      }
   }
}


static inline nNIKAL100_tU32 nNIKAL100_mapHashAddress(nNIKAL100_tUPtr key)
{ return ((key >> PAGE_SHIFT) % nNIKAL100_tMapHashSize); }



static nNIKAL100_tStatus nNIKAL100_mapInsert(nNIKAL100_tMap *map, nNIKAL100_tUPtr key, nNIKAL100_tUPtr value)
{
   nNIKAL100_tListNode *listIterator;
   nNIKAL100_tMapBufferNode *bufferPtr;
   nNIKAL100_tU32 hashValue = nNIKAL100_mapHashAddress(key);

   
   listIterator = ((*map)[hashValue]).next;
   for (;listIterator != &((*map)[hashValue]); listIterator = listIterator->next)
   {
      
      bufferPtr = (nNIKAL100_tMapBufferNode*)listIterator;
      if (bufferPtr->size == nNIKAL100_tMapBufferCapacity)
         continue;

      bufferPtr->keyValuePairs[bufferPtr->size].key = key;
      bufferPtr->keyValuePairs[bufferPtr->size].value = value;
      ++(bufferPtr->size);

      
      return nNIKAL100_kStatusSuccess;
   }

   
   
   bufferPtr = (nNIKAL100_tMapBufferNode*)kmalloc(
      sizeof(nNIKAL100_tMapBufferNode), GFP_KERNEL);
   if (bufferPtr == NULL)
   {
      nNIKAL100_debugPrintf("nNIKAL100_mapInsert: Failed to allocate bufNode: map: %p, key: %lx(hash:%lx), value: %lx\n", map, key, (unsigned long)hashValue, value);
      return nNIKAL100_kStatusMemoryFull;
   }

   nNIKAL100_listAdd(&((*map)[hashValue]), &(bufferPtr->listNode));
   bufferPtr->size = 1;
   bufferPtr->keyValuePairs[0].key = key;
   bufferPtr->keyValuePairs[0].value = value;

   
   return nNIKAL100_kStatusSuccess;
}


static void nNIKAL100_mapFind(nNIKAL100_tMap *map, nNIKAL100_tUPtr key, nNIKAL100_tMapIterator *iterator)
{
   nNIKAL100_tListNode *listIterator;
   nNIKAL100_tMapBufferNode *bufferPtr;
   nNIKAL100_tU32 hashValue = nNIKAL100_mapHashAddress(key);
   nNIKAL100_tU32 i;

   listIterator = ((*map)[hashValue]).next;
   for (;listIterator != &((*map)[hashValue]); listIterator = listIterator->next)
   {
      
      bufferPtr = (nNIKAL100_tMapBufferNode*)listIterator;

      for (i=0; i<bufferPtr->size; ++i)
      {
         if (key == bufferPtr->keyValuePairs[i].key)
         {
            
            iterator->mapPtr = map;
            iterator->bucketPtr = &((*map)[hashValue]);
            iterator->bufferPtr = bufferPtr;
            iterator->keyValuePtr = &(bufferPtr->keyValuePairs[i]);

            
            return;
         }
      }
   }

   
   iterator->mapPtr = NULL;
   iterator->bucketPtr = NULL;
   iterator->bufferPtr = NULL;
   iterator->keyValuePtr = NULL;
   return;
}


static void nNIKAL100_mapRemove(nNIKAL100_tMap *map, const nNIKAL100_tMapIterator *iterator)
{
   nNIKAL100_tU32 numberOfEntriesToShiftUp;
   nNIKAL100_tMapBufferNode *bufferPtr = iterator->bufferPtr;
   nNIKAL100_tMapKeyValuePair *keyValuePtr = iterator->keyValuePtr;

   if (iterator->mapPtr == NULL)
   {
      nNIKAL100_debugPrintf("nNIKAL100_mapRemove: iterator not valid: mapPtr: %p, bucketPtr: %p, bufferPtr: %p, keyValuePtr: %p\n",
         iterator->mapPtr, iterator->bucketPtr, iterator->bufferPtr, iterator->keyValuePtr);
      return;
   }

   

   
   nNIKAL100_assertFatal(bufferPtr->size > 0, "nNIKAL100_mapRemove: trying to remove from 0 sized buffer\n");
   --(bufferPtr->size);
   numberOfEntriesToShiftUp = bufferPtr->size - (iterator->keyValuePtr - &(bufferPtr->keyValuePairs[0]));
   nNIKAL100_assertFatal(numberOfEntriesToShiftUp <= bufferPtr->size, "nNIKAL100_mapRemove: trying to shift too many entries during remove.\n");

   for (; numberOfEntriesToShiftUp>0; --numberOfEntriesToShiftUp)
   {
      *keyValuePtr = *(keyValuePtr+1);
      ++keyValuePtr;
   }

   
   if (bufferPtr->size == 0)
   {
      nNIKAL100_listRemove(iterator->bucketPtr, &(bufferPtr->listNode));
      
      kfree(bufferPtr);
   }
}


static inline void nNIKAL100_pageAlignMemoryBlock(void *inPtr, nNIKAL100_tUPtr size, void **pageAlignedPtr,
   nNIKAL100_tUPtr *pageAlignedSize, nNIKAL100_tUPtr *offset)
{
   nNIKAL100_tUPtr pageSize, addr, outAddr, endAddr;
   pageSize = nNIKAL100_getPageSize();
   addr = (nNIKAL100_tUPtr)inPtr;
   *offset = addr & (pageSize-1);
   outAddr = addr & (~(pageSize-1));
   endAddr = ((addr+size) + (pageSize-1)) & (~(pageSize-1));
   *pageAlignedSize = endAddr - outAddr;
   *pageAlignedPtr = (void*)outAddr;
}

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_registerDriver(nNIKAL100_tDriver *driver)
{
   nNIKAL100_tListNode *i;
   nNIKAL100_tStatus status = nNIKAL100_kStatusSuccess;

   down(&(nNIKAL100_driverList.lock));

   
   for (i = nNIKAL100_driverList.head.next; i != &(nNIKAL100_driverList.head);
      i = i->next)
   {
      if (((nNIKAL100_tDriver*)i)->number == driver->number)
         status = nNIKAL100_kStatusResourceBusy;
   }

   if (nNIKAL100_statusIsNonfatal(status))
      nNIKAL100_listAdd(&(nNIKAL100_driverList.head), &(driver->listNode));
   up(&(nNIKAL100_driverList.lock));

   if (nNIKAL100_statusIsFatal(status))
      return status;

#if nNIKAL100_mIsLinux24
   if (nNIKAL100_statusIsNonfatal(status))
   {
      sNipalkDevFSHandle = devfs_register(NULL, driver->name, DEVFS_FL_DEFAULT, nNIKAL100_gDeviceMajor, driver->number,
            S_IFCHR | S_IRUGO | S_IWUGO,
            &nNIKAL100_fops, NULL);

      
      return nNIKAL100_kStatusSuccess;
   }
#else
   return nNIKAL100_kStatusSuccess;
#endif

   
   nNIKAL100_unregisterDriver(driver);

   return status;
}

nNIKAL100_cc void nNIKAL100_unregisterDriver(nNIKAL100_tDriver *driver)
{
#if nNIKAL100_mIsLinux24
   devfs_unregister(sNipalkDevFSHandle);
#endif

   down(&(nNIKAL100_driverList.lock));
   nNIKAL100_listRemove(&(nNIKAL100_driverList.head), &(driver->listNode));
   up(&(nNIKAL100_driverList.lock));

   driver->unload(driver);
}


static int __init nNIKAL100_initDriver(void)
{
   int status = 0;

   /** We don't support machine configuration with physical memory located
    *  above the 4GB physical address boundary on 32bit systems.
    *  This typically occurs for systems that have more than 3GB memory.  The
    *  BIOS needs to reserve some physical address space for device mapping,
    *  so it puts any additional memory beyond 3GB above the 4GB physical
    *  address space.
    *  Below, we check if addressableMemory overflows a 32 bit value.
    *  On 32bit systems this occurs when there is more than 4GB of addressable memory.
    **/
#ifdef __i386__
   nNIKAL100_tU64 addressableMemory;

   nNIKAL100_assertFatal( max_mapnr != 0, "nNIKAL100_initDriver; max_mapnr == 0!\n" );
   nNIKAL100_assertFatal( PAGE_SIZE != 0, "nNIKAL100_initDriver; PAGE_SIZE == 0!\n" );

   addressableMemory = ((nNIKAL100_tU64)(max_mapnr) * (nNIKAL100_tU64)PAGE_SIZE) - 1;

   
   if ( addressableMemory >> 32 )
   {
      printk(KERN_WARNING "[nikal] More than 4GB of addressable memory detected.\n");
      printk(KERN_WARNING "[nikal] This configuration is not supported.  Check the release notes for more information.\n");
      return -ENOSYS;
   }
#endif

   
#if nNIKAL100_mIsLinux24
   nNIKAL120_sStackSize = nNIKAL120_mRoundUp(((nNIKAL100_tUPtr)(&status) -
         (nNIKAL100_tUPtr)(current)), PAGE_SIZE) - sizeof(nLinux_task);
#else
   nNIKAL120_sStackSize = nNIKAL120_mRoundUp(((nNIKAL100_tUPtr)(&status) -
         (nNIKAL100_tUPtr)(current_thread_info())), PAGE_SIZE) -
         sizeof(struct thread_info);
#endif
   nNIKAL100_debugPrintf("Kernel stack size: %lu\n", nNIKAL120_sStackSize);

   
   nNIKAL100_listInit(&(nNIKAL100_driverList.head));
   sema_init(&(nNIKAL100_driverList.lock), 1);
   sema_init(&(nNIKAL100_sMMapSema), 1);
   sema_init(&(nNIKAL100_sPageLockOperationSema), 1);
   nNIKAL100_mapInit(&nNIKAL100_sPageLockMap);
   nNIKAL100_mapInit(&nNIKAL100_sMallocContiguousMap);
   sema_init(&(nNIKAL100_sMallocContiguousMapLock), 1);

#if nNIKAL100_mIsLinux24
   nNIKAL110_sJiffiesLastRead = jiffies;
   spin_lock_init(&nNIKAL110_sJiffiesRolloverSpinlock);

   SET_MODULE_OWNER(&nNIKAL100_fops);
   if (nNIKAL100_gDeviceMajor)
      status = devfs_register_chrdev(nNIKAL100_gDeviceMajor, "nikal", &nNIKAL100_fops);
   else
   {
      nNIKAL100_gDeviceMajor = devfs_register_chrdev(nNIKAL100_gDeviceMajor, "nikal", &nNIKAL100_fops);
      status = ((nNIKAL100_gDeviceMajor<0)? nNIKAL100_gDeviceMajor : 0);
   }
#else
#ifndef HAVE_COMPAT_IOCTL
   status = register_ioctl32_conversion(0xc018d501, NULL);
   if (status < 0)
   {
      return status;
   }
#endif
   cdev_init(&sNipalkCDev, &nNIKAL100_fops);
   sNipalkCDev.owner = THIS_MODULE;
   kobject_set_name(&(sNipalkCDev.kobj), "nikal");

   if (nNIKAL100_gDeviceMajor)
   {
      sNipalkDeviceNumber = MKDEV(nNIKAL100_gDeviceMajor, 0);
      status = register_chrdev_region(sNipalkDeviceNumber, 256, "nikal");
   }
   else
   {
      status = alloc_chrdev_region(&sNipalkDeviceNumber, 0, 256, "nikal");
      nNIKAL100_gDeviceMajor = MAJOR(sNipalkDeviceNumber);
   }

   if (status >= 0)
   {
      status = cdev_add(&sNipalkCDev, sNipalkDeviceNumber, 256);

      if (status >= 0)
      {
         return 0; 
      }

      kobject_put(&(sNipalkCDev.kobj));
   }
#endif

   return status;
}


static void __exit nNIKAL100_cleanupDriver(void)
{
   nNIKAL100_debugPrintf("Max stack usage seen: %lu\n", nNIKAL120_sMaxStackUsageSeen);
#if nNIKAL100_mIsLinux24
   devfs_unregister_chrdev(nNIKAL100_gDeviceMajor, "nikal");;
#else
#ifndef HAVE_COMPAT_IOCTL
   unregister_ioctl32_conversion(0xc018d501);
#endif
   cdev_del(&sNipalkCDev);

   unregister_chrdev_region(sNipalkDeviceNumber, 256);
#endif

   nNIKAL100_mapCleanup(&nNIKAL100_sMallocContiguousMap);
   nNIKAL100_mapCleanup(&nNIKAL100_sPageLockMap);
}


static int nNIKAL100_open(nLinux_inode *the_inode, nLinux_fileHandle *filePtr)
{
   int status = 0;
   nNIKAL100_tFilePrivateData *filePrivateData;
   nNIKAL100_tListNode *i;
   nNIKAL100_tStatus kalStatus = nNIKAL100_kStatusSuccess;

   nNIKAL100_MOD_INC_USE_COUNT;

   filePrivateData = (nNIKAL100_tFilePrivateData*)nNIKAL100_malloc(sizeof(nNIKAL100_tFilePrivateData));
   if (filePrivateData != NULL)
   {
      
      filePrivateData->driver = NULL;
      down(&(nNIKAL100_driverList.lock));
      for (i = nNIKAL100_driverList.head.next; i != &(nNIKAL100_driverList.head);
         i = i->next)
      {
         if (((nNIKAL100_tDriver*)i)->number == MINOR(the_inode->i_rdev))
         {
            filePrivateData->driver = (nNIKAL100_tDriver*)i;
            break;
         }
      }
      up(&(nNIKAL100_driverList.lock));

      if (filePrivateData->driver != NULL)
      {
         if (filePrivateData->driver->module)
            kalStatus = nNIKAL100_incrementDriverRefcount(filePrivateData->driver->module);

         if (nNIKAL100_statusIsNonfatal(kalStatus))
         {
            kalStatus = filePrivateData->driver->open(filePtr, &(filePrivateData->context));
            filePtr->private_data = filePrivateData;
            if (nNIKAL100_statusIsNonfatal(kalStatus))
               return 0;
            else
               status = -ENODEV;

            if (filePrivateData->driver->module)
               nNIKAL100_decrementDriverRefcount(filePrivateData->driver->module);
         }
         else
            status = -ENODEV;
      }
      else
         status = -ENODEV;

      nNIKAL100_free(filePrivateData);
   }
   else
      status = -ENOMEM;

   nNIKAL100_MOD_DEC_USE_COUNT;
   return status;
}


static int nNIKAL100_release(nLinux_inode *the_inode, nLinux_fileHandle *filePtr)
{
   nNIKAL100_tFilePrivateData *filePrivateData =
      (nNIKAL100_tFilePrivateData*)filePtr->private_data;

   if (filePrivateData->driver->close)
      filePrivateData->driver->close(filePrivateData->context);
   if (filePrivateData->driver->module)
      nNIKAL100_decrementDriverRefcount(filePrivateData->driver->module);
   nNIKAL100_MOD_DEC_USE_COUNT;
   return 0;
}


static int nNIKAL100_ioctl(nLinux_inode *the_inode, nLinux_fileHandle *filePtr,
   unsigned int command, unsigned long param)
{
   int status = -EINVAL;
   nNIKAL100_tFilePrivateData *filePrivateData =
      (nNIKAL100_tFilePrivateData*)filePtr->private_data;

   if (filePrivateData->driver->deviceIOControl)
   {
      nNIKAL100_tStatus kalStatus =
         filePrivateData->driver->deviceIOControl(filePrivateData->context, command, (void*)param);
      
      if (nNIKAL100_statusIsFatal(kalStatus))
         status = -ENOTTY;
      else
         status = 0;
   }

   return status;
}
#ifdef HAVE_COMPAT_IOCTL
static long nNIKAL100_compatIoctl(nLinux_fileHandle *filePtr, unsigned int command,
   unsigned long param)
{
   return (long)nNIKAL100_ioctl(NULL, filePtr, command, param);
}
#endif


static ssize_t nNIKAL100_read(nLinux_fileHandle *filePtr, char __user *buffer, size_t count, loff_t *offset)
{
   ssize_t ret = -EINVAL;
   nNIKAL100_tIPtr _offset = *offset;
   nNIKAL100_tIPtr bytesRead;
   nNIKAL100_tFilePrivateData *filePrivateData =
      (nNIKAL100_tFilePrivateData*)filePtr->private_data;

   if (filePrivateData->driver->read)
   {
      nNIKAL100_tStatus status;
      
      char * readBuffer = nNIKAL100_malloc(count);
      if (readBuffer == NULL)
      {
         return -ENOMEM;
      }

      status = filePrivateData->driver->read(filePrivateData->context, readBuffer, count, &_offset, &bytesRead);

      if (nNIKAL100_statusIsNonfatal(status))
         ret = 0;
      else
         ret = -EINVAL;

      nNIKAL100_replicateOutboundIOControlBuffer(buffer, readBuffer, bytesRead);
      nNIKAL100_free(readBuffer);
   }

   *offset = _offset;

   if (ret < 0)
      return ret;
   else
      return bytesRead;
}


static ssize_t nNIKAL100_write(nLinux_fileHandle *filePtr, const char __user *buffer, size_t count, loff_t *offset)
{
   ssize_t ret = -EINVAL;
   nNIKAL100_tIPtr _offset = *offset;
   nNIKAL100_tIPtr bytesWritten;
   nNIKAL100_tFilePrivateData *filePrivateData =
      (nNIKAL100_tFilePrivateData*)filePtr->private_data;

   if (filePrivateData->driver->write)
   {
      nNIKAL100_tStatus status;
      
      char * writeBuffer = nNIKAL100_malloc(count);
      if (writeBuffer == NULL)
      {
         return -ENOMEM;
      }
      status = nNIKAL100_replicateInboundIOControlBuffer(writeBuffer, buffer, count);
      if (nNIKAL100_statusIsNonfatal(status))
      {
         status = filePrivateData->driver->write(filePrivateData->context, writeBuffer, count, &_offset, &bytesWritten);

         if (nNIKAL100_statusIsNonfatal(status))
            ret = 0;
         else
            ret = -EINVAL;

         nNIKAL100_free(writeBuffer);
      }
      else
      {
         return -EINVAL;
      }
   }


   *offset = _offset;

   if (ret < 0)
      return ret;
   else
      return bytesWritten;
}


static int nNIKAL100_mmap(nLinux_fileHandle *filePtr, nLinux_vmArea *vma)
{
   int status = -EINVAL;
   nNIKAL100_tUPtr address;
   nNIKAL100_tUPtr physAddr;
   u32 pageSize;
   nNIKAL100_tBoolean isFromIo;
   nNIKAL100_tStatus remapStatus;

   
   if (nNIKAL100_gMMapDataPtr->callingThreadID != current->pid)
   {
      
      status = -EPERM;
   }
   else
   {
      nNIKAL100_assertFatal(nNIKAL100_gMMapDataPtr->addressType != nNIKAL100_kMemoryAddressTypeInvalid,
         "addressType passed to nNIKAL100_privateMMap is invalid!\n");

      
      pageSize = nNIKAL100_getPageSize();
      address = vma->vm_pgoff * pageSize;
      physAddr = 0;
      if (nNIKAL100_gMMapDataPtr->addressType == nNIKAL100_kMemoryAddressTypeVirtual)
      {
         nNIKAL100_tUPtr i;
         i = 0;
         nNIKAL100_compileTimeAssert(sizeof(void*)==sizeof(nNIKAL100_tUPtr), "Oops, resizing void* into nNIKAL100_tUPtr!");
         physAddr = nNIKAL100_getPhysicalAddress((void*)address, nNIKAL100_kTrue);
         isFromIo = !pfn_valid(physAddr >> PAGE_SHIFT);

         
         #ifdef nNIKAL100_kDebuggingIsActive
         pageSize = nNIKAL100_getPageSize();
         for (i=address+pageSize; i < address + ((vma->vm_end) - (vma->vm_start)); i+=pageSize)
         {
            nNIKAL100_assertFatal(isFromIo !=
               pfn_valid(nNIKAL100_getPhysicalAddress((void*)i, nNIKAL100_kTrue) >>
                  PAGE_SHIFT),
               "nNIKAL100_mmap() asked to map a region of non-uniform memory\n");
         }
         #endif
      }
      else 
      {
         physAddr = address;
         isFromIo = nNIKAL100_kTrue;
      }

      if (isFromIo)
      {
         remapStatus = nNIKAL100_remapPageRange(vma, physAddr);
         nNIKAL100_debugPrintf("nNIKAL100_mmap() From I/O, address: %lx, phys: %lx, size: %lu, ioMemory, remap_page_range status: %ld\n", address, physAddr, vma->vm_end - vma->vm_start, (long)remapStatus);
         if (nNIKAL100_statusIsNonfatal(remapStatus))
         {
            vma->vm_flags |= VM_IO;
            status = 0;
         }
         else
         {
            status = -ENOMEM;
         }
      }
      else
      {
           
         vma->vm_ops = &nNIKAL100_vmaOps;
         nNIKAL100_vmaOpen(vma);
         
         status = 0;
      }
   }

   if (status >= 0)
   {
      vma->vm_flags |= VM_RESERVED | VM_LOCKED | VM_DONTCOPY | VM_DONTEXPAND;
   }

   return status;
}


static loff_t nNIKAL100_llseek(nLinux_fileHandle *filePtr, loff_t inOffset, int seekFrom)
{
   return -ESPIPE;
}


static void nNIKAL100_vmaOpen(nLinux_vmArea *vma)
{ nNIKAL100_MOD_INC_USE_COUNT; }


static inline nNIKAL100_tBoolean nNIKAL100_isKernelContiguousPointer(const void *ptr)
{
   return ((nNIKAL100_tUPtr)ptr < (nNIKAL100_tUPtr)VMALLOC_START);
}


static inline nNIKAL100_tBoolean nNIKAL100_isKernelVirtualPointer(const void *ptr)
{
   return ((((nNIKAL100_tUPtr)ptr >= (nNIKAL100_tUPtr)VMALLOC_START) &&
      ((nNIKAL100_tUPtr)ptr < (nNIKAL100_tUPtr)VMALLOC_END)));
}


static void nNIKAL100_vmaClose(nLinux_vmArea *vma)
{
   nNIKAL100_tUPtr i; 
   nNIKAL100_tUPtr mappedKernelAddress = (vma->vm_pgoff << PAGE_SHIFT);
   nNIKAL100_tUPtr size = vma->vm_end - vma->vm_start;
   nNIKAL100_tU32 pageSize;

   nNIKAL100_MOD_DEC_USE_COUNT;

   
   if (nNIKAL100_isKernelContiguousPointer((void*)mappedKernelAddress))
   {
      
      pageSize = nNIKAL100_getPageSize();
      for (i = mappedKernelAddress; i < mappedKernelAddress+size; i+=pageSize)
      {
         nNIKAL100_decrementPageRefcount((void*)i);
      }
   }

}

#ifdef nNIKAL190_kVmFaultMethod

static int nNIKAL190_vmaPageFaultHandler(nLinux_vmArea *vma, nLinux_vmFault *vmf)
{
   nNIKAL100_tUPtr addr = vmf->pgoff << PAGE_SHIFT;
   vmf->page = (nLinux_physicalPage*)nNIKAL100_incrementPageRefcount((void*)addr);
   return VM_CAN_NONLINEAR;
}
#else

#if nNIKAL100_mIsLinux24
static nLinux_physicalPage *nNIKAL100_vmaPageFaultHandler(nLinux_vmArea *vma,
      unsigned long address, int write_access)
#else
static nLinux_physicalPage *nNIKAL100_vmaPageFaultHandler(nLinux_vmArea *vma,
      unsigned long address, int *type)
#endif
{
   nNIKAL100_tUPtr addr = (address - vma->vm_start) + (vma->vm_pgoff << PAGE_SHIFT);
   nLinux_physicalPage *page = (nLinux_physicalPage*)nNIKAL100_incrementPageRefcount((void*)addr);
   return page;
}
#endif


static nLinux_vmArea *nNIKAL100_getVMA(nNIKAL100_tUPtr addr)
{
   nLinux_vmArea *vma = find_vma(current->mm, addr);

   
   if ((vma != NULL) && (addr >= vma->vm_start) && (addr < vma->vm_end))
      return vma;
   else
      return NULL;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_replicateOutboundIOControlBuffer(void __user *to, const void *from, nNIKAL100_tI32 count)
{
   if (copy_to_user(to, from, count))
      return nNIKAL100_kStatusOSFault;
   else
      return nNIKAL100_kStatusSuccess;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_replicateInboundIOControlBuffer(void *to, const void __user *from, nNIKAL100_tI32 count)
{
   if (copy_from_user(to, from, count))
      return nNIKAL100_kStatusOSFault;
   else
      return nNIKAL100_kStatusSuccess;
}


nNIKAL100_cc void nNIKAL100_getKernelVersion(nNIKAL100_tU32 *major, nNIKAL100_tU32 *minor)
{
   nNIKAL100_tU32 linVer = LINUX_VERSION_CODE;
   *major = (linVer >> 16);
   *minor = (linVer&0xFFFF) >> 8;
}


nNIKAL100_cc nNIKAL100_tI32 nNIKAL100_vsprintf(nNIKAL100_tText* buf, const nNIKAL100_tText* fmt, va_list arg)
{
   nNIKAL120_mCheckStackUsage;
   return vsprintf(buf, fmt, arg);
}


nNIKAL100_cc nNIKAL100_tI32 nNIKAL100_vsnPrintf(nNIKAL100_tText* buf, nNIKAL100_tU32 size, const nNIKAL100_tText* fmt, va_list arg)
{
   nNIKAL120_mCheckStackUsage;
   return vsnprintf(buf, size, fmt, arg);
}


nNIKAL100_cc void nNIKAL100_printToDebugger(const nNIKAL100_tText* string)
{
   nNIKAL120_mCheckStackUsage;
   printk("%s", string);
}


static nNIKAL100_tU32 nNIKAL100_convertSizeToPageOrder(nNIKAL100_tUPtr size)
{
   nNIKAL100_tU32 pageOrder;
   for (pageOrder = 0; (1 << pageOrder) < ((size + PAGE_SIZE - 1) / PAGE_SIZE); pageOrder++)
      ;
   return pageOrder;
}


nNIKAL100_cc void *nNIKAL100_malloc(nNIKAL100_tUPtr size)
{
   void *ptr = NULL;

   nNIKAL120_mCheckStackUsage;

   
   if (size < ( nNIKAL100_kKMallocMaxSize )) 
   {
      ptr = kmalloc(size, GFP_KERNEL);
   }
   if (ptr == NULL)
   {
      ptr = vmalloc(size);
   }

   return ptr;
}


nNIKAL100_cc void *nNIKAL100_mallocContiguous(nNIKAL100_tUPtr size)
{
   void *ptr = NULL;
   nNIKAL100_tU32 pageOrder;
   nLinux_physicalPage * page;
   nNIKAL100_tStatus status = nNIKAL100_kStatusSuccess;

   nNIKAL120_mCheckStackUsage;

   if (size < ( nNIKAL100_kKMallocMaxSize ))
   {
      ptr = kmalloc(size, GFP_KERNEL);
   }

   if (ptr == NULL)
   {
      pageOrder = nNIKAL100_convertSizeToPageOrder(size);
      page = alloc_pages(GFP_KERNEL,pageOrder);

      if (!page)
         return NULL;

      ptr = page_address(page);

      nNIKAL100_assertFatal((ptr == NULL) || nNIKAL100_isKernelContiguousPointer(ptr), "nNIKAL100_mallocContiguous allocated a non contiguous pointer.\n");

      
      down(&nNIKAL100_sMallocContiguousMapLock);
      status = nNIKAL100_mapInsert(&nNIKAL100_sMallocContiguousMap, (nNIKAL100_tUPtr)ptr, size);
      up(&nNIKAL100_sMallocContiguousMapLock);

      if (nNIKAL100_statusIsFatal(status))
      {
         free_pages((unsigned long)ptr,pageOrder);
         ptr = NULL;
      }
   }

   return ptr;
}


nNIKAL100_cc void *nNIKAL100_malloc32BitPhysicalContiguous(nNIKAL100_tUPtr size)
{
#if nNIKAL100_mIsLinux24
   return NULL;
#else
   void *ptr = NULL;

   dma_addr_t dma_handle;
   int headerSize;
#ifdef nNIKAL100_kDebuggingIsActive
   nNIKAL100_tUPtr address;
   u32 pageSize;
   nNIKAL100_tUPtr i;
   nNIKAL100_tU64 physAddr;
#endif

   nNIKAL120_mCheckStackUsage;

   headerSize = (sizeof(nNIKAL100_tUPtr) + sizeof(dma_addr_t) + 7) & -8;
   size = size + headerSize;
   ptr = dma_alloc_coherent(NULL, size, &dma_handle, GFP_KERNEL);

   if (ptr == NULL)
      return NULL;

   
   *(nNIKAL100_tUPtr *)ptr = size;
   *(dma_addr_t *)((nNIKAL100_tUPtr *)ptr + 1) = dma_handle;

#ifdef nNIKAL100_kDebuggingIsActive
   address = (nNIKAL100_tUPtr)ptr;
   pageSize = nNIKAL100_getPageSize();
   for (i=address; i < address + size; i+=pageSize)
   {
      physAddr = nNIKAL100_getPhysicalAddress((void*)i, nNIKAL100_kAddressOriginKernel);

      if ( physAddr >> 32 != 0 )
         nNIKAL100_debugPrintf("Not 32 bit physical! virtualAddr: %lx physAddr: %llx\n",
                               i, (long long)physAddr);

      nNIKAL100_assertFatal( physAddr >> 32 == 0, "nNIKAL100_malloc32BitPhysical; Not 32 bit Physical!\n" );
   }
#endif

   return (void *)((nNIKAL100_tUPtr)ptr + headerSize);
#endif
}


nNIKAL100_cc void nNIKAL100_free(void *pointer)
{
   nNIKAL100_tU32 pageOrder;
   nNIKAL100_tUPtr size = 0;
   nNIKAL100_tMapIterator mapIterator;

   nNIKAL120_mCheckStackUsage;

   if (!pointer)
      return;

   if (nNIKAL100_isKernelVirtualPointer(pointer))
   {
      vfree(pointer);
   }
   else
   {
      
      down(&nNIKAL100_sMallocContiguousMapLock);
      nNIKAL100_mapFind(&nNIKAL100_sMallocContiguousMap, (nNIKAL100_tUPtr)pointer, &mapIterator);
      if (mapIterator.keyValuePtr != NULL)
      {
         nNIKAL100_assertFatal(mapIterator.keyValuePtr->value != 0, "nNIKAL100_free(): found alloc_pages record but size == 0\n");
         size = mapIterator.keyValuePtr->value;
         nNIKAL100_mapRemove(&nNIKAL100_sMallocContiguousMap, &mapIterator);
      }
      up(&nNIKAL100_sMallocContiguousMapLock);

      if (size == 0)
      {
         kfree(pointer);
      }
      else
      {
         pageOrder = nNIKAL100_convertSizeToPageOrder(size);
         free_pages((unsigned long)pointer,pageOrder);
      }
   }
}


nNIKAL100_cc void nNIKAL100_free32BitPhysicalContiguous(void *pointer)
{
#if !nNIKAL100_mIsLinux24
   nNIKAL100_tUPtr size;
   dma_addr_t dma_handle;
   int headerSize;

   if ( pointer == NULL)
      return;

   headerSize = (sizeof(nNIKAL100_tUPtr) + sizeof(dma_addr_t) + 7) & -8;
   pointer = (void *)((nNIKAL100_tUPtr)pointer - headerSize);
   size = *(nNIKAL100_tUPtr *)pointer;
   dma_handle = *(dma_addr_t *)((nNIKAL100_tUPtr *)pointer + 1);

   dma_free_coherent(NULL, size, pointer, dma_handle);
#endif
}


nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_getPageSize(void)
{
   return PAGE_SIZE;
}


nNIKAL100_cc nNIKAL100_tUPtr nNIKAL100_getPhysicalMemorySize(void)
{
   nNIKAL100_tU64 memorySize;
   nLinux_sysinfo systemInfo;

   nNIKAL100_assertFatal( PAGE_SIZE != 0, "nNIKAL100_getPhysicalMemorySize; PAGE_SIZE == 0!\n" );
   si_meminfo(&systemInfo);
   
   nNIKAL100_assertFatal( systemInfo.totalram != 0, "nNIKAL100_getPhysicalMemorySize; systemInfo.totalram == 0!\n" );
   memorySize = (nNIKAL100_tU64)systemInfo.totalram * (nNIKAL100_tU64)PAGE_SIZE;

#ifdef __i386__
   
   nNIKAL100_assertFatal( memorySize >> 32 == 0, "nNIKAL100_getPhysicalMemorySize; More than 4GB of RAM!\n" );
#endif

   return (nNIKAL100_tUPtr)memorySize;
}


nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_getNumberOfActiveProcessors(void)
{
#if nNIKAL100_mIsLinux24
   return smp_num_cpus;
#else
   return num_online_cpus();
#endif
}

#if nNIKAL100_mIsLinux24

static inline nNIKAL100_tStatus nNIKAL100_allocKiovec(int nr, nLinux_kiobuf **iobuf)
{
#ifdef nNIKAL100_kAllocKiovecSZ
   int kioMaxSectors = KIO_MAX_SECTORS;

   if (alloc_kiovec_sz(nr, iobuf, &kioMaxSectors))
#else
   if (alloc_kiovec(nr, iobuf))
#endif
      return nNIKAL100_kStatusMemoryFull;
   else
      return nNIKAL100_kStatusSuccess;
}


static inline void nNIKAL100_freeKiovec(int nr, nLinux_kiobuf **iobuf)
{
#ifdef nNIKAL100_kAllocKiovecSZ
   int kioMaxSectors = KIO_MAX_SECTORS;

   free_kiovec_sz(nr, iobuf, &kioMaxSectors);
#else
   free_kiovec(nr, iobuf);
#endif
}
#endif


static void *nNIKAL100_mapUserKIOBuf(void *start, nNIKAL100_tUPtr size, nNIKAL100_tStatus *status)
{
#if !nNIKAL100_mIsLinux24
   const nNIKAL100_tI32 sizeInPages = (size/PAGE_SIZE);
   
   nLinux_physicalPage **buffer;
   nLinux_physicalPage **pages;
   nNIKAL100_tI32 pagesMapped;

   nNIKAL100_assertFatal(status != NULL, "NULL status passed to internal function\n");
   if (nNIKAL100_statusIsFatal(*status))
      return NULL;

   if (sizeInPages == 0)
   {
      nNIKAL100_debugPrintf("nNIKAL100_mapUserKIOBuf(): sizeInPages == 0\n");
      *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusValueConflict);
      return NULL;
   }

   
   if (current->mm == NULL)
   {
      nNIKAL100_debugPrintf("nNIKAL100_mapUserKIOBuf(): called when user mode address space is already destroyed\n");
      *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusResourceNotAvailable);
      return NULL;
   }

   buffer = (nLinux_physicalPage **)nNIKAL100_malloc(sizeof(nLinux_physicalPage **) * (sizeInPages+1));
   pages = buffer+1; 
   if (buffer == NULL)
   {
      nNIKAL100_debugPrintf("nNIKAL100_mapUserKIOBuf(): Failed malloc(%lu)\n", sizeof(nLinux_physicalPage **) * (sizeInPages+1));
      *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusMemoryFull);
      return NULL;
   }

   down_read(&(current->mm->mmap_sem));
   pagesMapped = get_user_pages(current, current->mm, (unsigned long) start,
      sizeInPages, 1 , 0 , pages, NULL );
   up_read(&(current->mm->mmap_sem));

   if (pagesMapped < 0)
   {
      
      if (pagesMapped == -EFAULT)
      {
         down_read(&(current->mm->mmap_sem));
         pagesMapped = get_user_pages(current, current->mm, (unsigned long) start,
            sizeInPages, 0 , 0 , pages, NULL );
         up_read(&(current->mm->mmap_sem));
      }
      
      if (pagesMapped < 0)
      {
         nNIKAL100_debugPrintf("nNIKAL100_mapUserKIOBuf: FAILED to map read-only! pagesMapped = %ld sizeInPages = %ld\n", (long)pagesMapped, (long)sizeInPages);
         nNIKAL100_free(buffer);
         *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusBadPointer);
         return NULL;
      }
   }
   if (pagesMapped == sizeInPages)
   {
      *(nNIKAL100_tUPtr*)buffer = sizeInPages;
      return buffer;
   }

   nNIKAL100_debugPrintf("nNIKAL100_mapUserKIOBuf: FAILED to map r/w! pagesMapped = %ld sizeInPages = %ld\n", (long)pagesMapped, (long)sizeInPages);
   
   for (;pagesMapped; --pagesMapped)
   {
      if (!PageReserved(pages[pagesMapped-1]))
      {
         page_cache_release(pages[pagesMapped-1]);
      }
   }

   nNIKAL100_free(buffer);
   *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusBadPointer);
   return NULL;
#else
   nLinux_kiobuf *iobuf;

   if (nNIKAL100_statusIsFatal(*status))
      return NULL;

   if ((nNIKAL100_allocKiovec(1, &iobuf)) != nNIKAL100_kStatusSuccess)
   {
      nNIKAL100_debugPrintf("nNIKAL100_mapUserKIOBuf(): nNIKAL100_allocKiovec() failed\n");
      *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusMemoryFull);
      return NULL;
   }

   
   if ((map_user_kiobuf(READ, iobuf, (unsigned long) start, size)) != 0)
   {
      
      if ((map_user_kiobuf(WRITE, iobuf, (unsigned long) start, size)) != 0)
      {
         
         nNIKAL100_debugPrintf("nNIKAL100_mapUserKIOBuf: map_user_kiobuf FAILED!\n");
         nNIKAL100_freeKiovec(1, &iobuf);
         *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusBadPointer);
         return NULL;
      }
   }
   return iobuf;
#endif
}


static void nNIKAL100_unmapUserKIOBuf(void *iobuf)
{
#if !nNIKAL100_mIsLinux24
   nLinux_physicalPage **buffer = (nLinux_physicalPage **)iobuf;
   nLinux_physicalPage **pages = buffer+1;
   nNIKAL100_tUPtr pagesMapped = *(nNIKAL100_tUPtr*)buffer;

   for (;pagesMapped; --pagesMapped)
   {
      if (!PageReserved(pages[pagesMapped-1]))
      {
         set_page_dirty(pages[pagesMapped-1]);
         
         page_cache_release(pages[pagesMapped-1]);
      }
   }
   nNIKAL100_free(buffer);
#else
#ifdef nNIKAL100_kMarkDirtyKIOBuf
   mark_dirty_kiobuf((nLinux_kiobuf*)iobuf, ((nLinux_kiobuf*)iobuf)->length );
#endif
   unmap_kiobuf((nLinux_kiobuf*)iobuf);
   nNIKAL100_freeKiovec(1, (nLinux_kiobuf**)(&iobuf));
#endif
}

static inline pmd_t *nNIKAL120_getPageTablePMDEntry(nLinux_mm* mm, const void *address)
{
   pgd_t *pgd;
#ifdef nNIKAL100_kFourLevelPageTable
   pud_t *pud;
#endif
   pmd_t *pmd;

   nNIKAL100_compileTimeAssert(sizeof(pte_t) <= 8, "pte_t is larger then 128 bits!\n");

   if (mm == &init_mm)
   {
      pgd = pgd_offset_k((nNIKAL100_tUPtr)address);
   }
   else
   {
      pgd = pgd_offset(mm, (nNIKAL100_tUPtr)address);
   }
#ifdef nNIKAL100_kFourLevelPageTable
   pud = pud_offset(pgd, (nNIKAL100_tUPtr)address);
   pmd = pmd_offset(pud, (nNIKAL100_tUPtr)address);
#else
   pmd = pmd_offset(pgd, (nNIKAL100_tUPtr)address);
#endif
   return pmd;
}


static inline pte_t *nNIKAL100_getKernelPageTableEntry(nLinux_mm* mm, void *address)
{
   pmd_t *pmd;
   pte_t *pte;
   pmd = nNIKAL120_getPageTablePMDEntry(mm, address);

#ifdef nNIKAL100_kPTEOffsetKernel
   pte = pte_offset_kernel(pmd, (nNIKAL100_tUPtr)address);
   return pte; 
#else
   pte = pte_offset(pmd, (nNIKAL100_tUPtr)address);
   return pte;
#endif
}


static inline pte_t *nNIKAL100_mapUserPageTableEntry(nLinux_mm* mm, void *address)
{
   pmd_t *pmd = nNIKAL120_getPageTablePMDEntry(mm, address);

#ifdef nNIKAL100_kPTEOffsetKernel
   return pte_offset_map(pmd, (nNIKAL100_tUPtr)address);
#else
   return pte_offset(pmd, (nNIKAL100_tUPtr)address);
#endif
}


static inline void nNIKAL100_unmapUserPageTableEntry(pte_t *pteToCleanup)
{
#ifdef nNIKAL100_kPTEOffsetKernel
   pte_unmap(pteToCleanup);
#endif
}


static void nNIKAL100_copyUserPageTable(void *userAddress, void *kernelAddress,
   nNIKAL100_tUPtr size, void *originalPageTableEntry)
{
   void *i, *j;
   nNIKAL100_tUPtr k;
   pte_t* pteKernel, *pteUser;

   size = (size + PAGE_SIZE - 1)/PAGE_SIZE;
   flush_cache_all();

   down_read(&(current->mm->mmap_sem));
   spin_lock(&init_mm.page_table_lock);

   *((pte_t*)originalPageTableEntry) = *nNIKAL100_getKernelPageTableEntry(&init_mm, kernelAddress);
   for (i=userAddress, j=kernelAddress, k=0; k<size;
      i+=PAGE_SIZE, j+=PAGE_SIZE, ++k)
   {
      pteKernel = nNIKAL100_getKernelPageTableEntry(&init_mm, j);
      pteUser = nNIKAL100_mapUserPageTableEntry(current->mm, i);
      *pteKernel = *pteUser;
      

      nNIKAL100_unmapUserPageTableEntry(pteUser);
   }
   spin_unlock(&init_mm.page_table_lock);
   up_read(&(current->mm->mmap_sem));
   flush_cache_all();
}


static void nNIKAL100_setKernelPageTable(void *newPte, void *kernelAddress, nNIKAL100_tUPtr size)
{
   void *i;
   nNIKAL100_tUPtr k;
   size = (size + PAGE_SIZE - 1)/PAGE_SIZE;

   flush_cache_all();
   spin_lock(&init_mm.page_table_lock);
   for (i=kernelAddress, k=0; k<size; i+=PAGE_SIZE, ++k)
      *nNIKAL100_getKernelPageTableEntry(&init_mm, i) = *((pte_t*)newPte);
   spin_unlock(&init_mm.page_table_lock);
   flush_cache_all();
}


static nNIKAL100_tStatus nNIKAL100_remapPageRange(void *vmaPtr, nNIKAL100_tUPtr physAddr)
{
   int status;
   nLinux_vmArea *vma = (nLinux_vmArea*)vmaPtr;
   vma->vm_flags |= (VM_RESERVED | VM_IO);
   
   status =
#ifdef nNIKAL100_kRemapPFNRange
   remap_pfn_range(vma, vma->vm_start, physAddr >> PAGE_SHIFT,
                   (vma->vm_end - vma->vm_start),
                   vma->vm_page_prot);
#else
#ifdef nNIKAL100_kFiveParameterRemapPageRange
   remap_page_range(vma, vma->vm_start, physAddr,
                    (vma->vm_end - vma->vm_start),
                    vma->vm_page_prot);
#else
   remap_page_range(vma->vm_start, physAddr,
                    (vma->vm_end - vma->vm_start),
                    vma->vm_page_prot);
#endif
#endif
   return nNIKAL100_convertLinuxToKALStatus(status);
}


static nNIKAL100_tUPtr nNIKAL100_doMMap(void *file, nNIKAL100_tUPtr address,
   nNIKAL100_tUPtr length, nNIKAL100_tUPtr offset)
{
   nNIKAL100_tUPtr mmapAddress;
   
   if (current->mm == NULL)
   {
      nNIKAL100_debugPrintf("nNIKAL100_doMMap(): called when user mode address space is already destroyed\n");
      return -1; 
   }
   down_write(&(current->mm->mmap_sem));
   mmapAddress = do_mmap((nLinux_fileHandle*)file, address, length, PROT_READ|PROT_WRITE,
      MAP_SHARED, offset);
   up_write(&(current->mm->mmap_sem));
   return mmapAddress;
}


static nNIKAL100_tStatus nNIKAL100_doMUnmap(nNIKAL100_tUPtr address, nNIKAL100_tUPtr length)
{
   int status = 0;
   nNIKAL100_tStatus kalStatus = nNIKAL100_kStatusSuccess;

   
   if (current->mm != NULL)
   {
      down_write(&(current->mm->mmap_sem));
      if (nNIKAL100_getVMA(address) != NULL)
      {
#ifdef nNIKAL100_kFourParameterDoMunmap
         status = do_munmap(current->mm, address, length, 1);
#else
         status = do_munmap(current->mm, address, length);
#endif
         kalStatus = nNIKAL100_convertLinuxToKALStatus(status);
         if (nNIKAL100_statusIsFatal(kalStatus))
         {
            nNIKAL100_debugPrintf("nNIKAL100_doMUnmap(), do_munmap(%lx, %lx) failed: status: %d(%ld)\n", address, length, status, (long)kalStatus);
         }
      }
      else
      {
         nNIKAL100_debugPrintf("nNIKAL100_doMUnmap(), can't find VMA, addr: %lx, size: %lx\n", address, length);
      }
      up_write(&(current->mm->mmap_sem));
   }
   else
   {
      nNIKAL100_debugPrintf("nNIKAL100_doMUnmap(), User address space destroyed, addr: %lx, size: %lx\n", address, length);
   }
   return kalStatus;
}


nNIKAL100_cc void *nNIKAL180_memCpy(void *dest, const void *src, size_t size)
{
   return memcpy(dest, src, size);
}


nNIKAL100_cc void *nNIKAL180_memMove(void *dest, const void *src, size_t size)
{
   return memmove(dest, src, size);
}


nNIKAL100_cc nNIKAL100_tUPtr nNIKAL100_getPhysicalAddress(const void *ptr, nNIKAL100_tAddressOrigin addressOrigin)
{
   pmd_t *_pmd;
   pte_t *_pte;
   nLinux_mm *_mm = &init_mm; 
   nLinux_physicalPage *_page = NULL;

   
   if (addressOrigin == nNIKAL100_kAddressOriginKernel)
   {
      if (nNIKAL100_isKernelContiguousPointer(ptr))
      {
         
         return __pa(ptr);
      }
   }
   else
   {
      
      _mm = current->mm;
      if (!_mm)
      {
         nNIKAL100_assertFatal(0, "nNIKAL100_getPhysicalAddress: user address space not found!\n");
         return 0;
      }
   }

   spin_lock(&(_mm->page_table_lock));
   _pmd = nNIKAL120_getPageTablePMDEntry(_mm, ptr);

   if (pmd_present(*_pmd))
   {
#ifdef nNIKAL100_kPTEOffsetKernel
      if (addressOrigin == nNIKAL100_kAddressOriginUser)
         _pte = pte_offset_map(_pmd, (nNIKAL100_tUPtr)ptr);
      else
      _pte = pte_offset_kernel(_pmd, (nNIKAL100_tUPtr)ptr);
#else
      _pte = pte_offset(_pmd, (nNIKAL100_tUPtr)ptr);
#endif
      if (pte_present(*_pte))
         _page = pte_page(*_pte);

#ifdef nNIKAL100_kPTEOffsetKernel
      if (addressOrigin == nNIKAL100_kAddressOriginUser)
         pte_unmap(_pte);
#endif
   }
   spin_unlock(&(_mm->page_table_lock));
   if (_page)
   {
      return (page_to_phys(_page) + ((nNIKAL100_tUPtr)ptr & (PAGE_SIZE-1)));
   }
   else
   {
      nNIKAL100_assertFatal(0, "nNIKAL100_getPhysicalAddress: Failed to walk the page table!\n");
      return 0;
   }
}


static void *nNIKAL100_incrementPageRefcount(const void *ptr)
{
   nNIKAL100_tUPtr physAddr = nNIKAL100_getPhysicalAddress(ptr, nNIKAL100_kAddressOriginKernel);
   nLinux_physicalPage *page = virt_to_page(__va(physAddr)); 
   get_page(page); 
   return page;
}


static void nNIKAL100_decrementPageRefcount(const void *ptr)
{
   nNIKAL100_tUPtr physAddr = nNIKAL100_getPhysicalAddress(ptr, nNIKAL100_kAddressOriginKernel);
   nLinux_physicalPage *page = virt_to_page(__va(physAddr)); 
   put_page_testzero(page); 
}


nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_getTimerInterval (void)
{
   return 1000000/HZ;
}




nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_sleepTimeoutInterruptible(
   nNIKAL100_tU32 timeToSleepInJiffies)
{
   set_current_state(TASK_INTERRUPTIBLE);
   return schedule_timeout(timeToSleepInJiffies);
}



nNIKAL100_cc void nNIKAL100_sleepTimeout (nNIKAL100_tU32 timeToSleepInJiffies)
{
   set_current_state(TASK_UNINTERRUPTIBLE);
   schedule_timeout(timeToSleepInJiffies);
   return;
}


nNIKAL100_cc nNIKAL100_tU64 nNIKAL110_getTimerCount (void)
{
#if nNIKAL100_mIsLinux24
   static nNIKAL100_tU32 nNIKAL110_sJiffiesRollovers = 0;
   nNIKAL100_tU64 kalJiffies64;

   spin_lock(&nNIKAL110_sJiffiesRolloverSpinlock);

   if (jiffies < nNIKAL110_sJiffiesLastRead)
      nNIKAL110_sJiffiesRollovers += 1;
   nNIKAL110_sJiffiesLastRead = jiffies;

   kalJiffies64 = nNIKAL110_sJiffiesRollovers;
   kalJiffies64 <<= 32;
   kalJiffies64 += nNIKAL110_sJiffiesLastRead;

   spin_unlock(&nNIKAL110_sJiffiesRolloverSpinlock);

   return kalJiffies64;
#else
   return get_jiffies_64();
#endif
}


nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_getTimerCount (void)
{
   return jiffies;
}


nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_getTimeOfDayInterval (void)
{
   const nNIKAL100_tU32 oneMicrosecond = 1;

#ifndef CONFIG_X86_TSC
   

   const nNIKAL100_tU32 microsecondsPerSecond = 1000000;
   nNIKAL100_tU32 clockTickInterval;

   clockTickInterval = microsecondsPerSecond/CLOCK_TICK_RATE;
   if (clockTickInterval == 0)
   {
      return oneMicrosecond;
   }
   else
   {
      return clockTickInterval;
   }

#else
   

   return oneMicrosecond;

#endif
}


nNIKAL100_cc void nNIKAL100_getTimeOfDay (nNIKAL100_tU32 *systemSeconds,
   nNIKAL100_tU32 *systemMicroSeconds)
{
   nLinux_timeValue systemTime;
   do_gettimeofday(&systemTime);
   *systemSeconds = systemTime.tv_sec;
   *systemMicroSeconds = systemTime.tv_usec;
}


nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_getThreadID( void )
{
   return (nNIKAL100_tU32)current->pid;
}


nNIKAL100_cc void *nNIKAL100_getThreadHandle( void )
{
   return (void *)current;
}


static int nNIKAL100_threadMain( void * paramBlock )
{
   nNIKAL100_tCreateThreadParamBlock *_paramBlock =
      (nNIKAL100_tCreateThreadParamBlock *)paramBlock;
   nNIKAL100_tThreadFunction _threadFunction = _paramBlock->threadFunction;
   void *_threadContext = _paramBlock->threadContext;

   if (atomic_dec_and_test(&(_paramBlock->counter)))
   {
      nNIKAL100_releaseSingleUseEvent( _paramBlock->threadCreatedEvent );
      _paramBlock->threadCreatedEvent = NULL;
   }

   return _threadFunction(_threadContext);
}
#ifndef nNIKAL150_kKthread
static void nNIKAL100_launchThread( void *params )
{
#ifndef nNIKAL160_kWorkqueueNonDelay
   nNIKAL100_tCreateThreadParamBlock *_paramBlock =
      (nNIKAL100_tCreateThreadParamBlock *)params;
#else
   nNIKAL100_tCreateThreadParamBlock *_paramBlock =
      container_of((nNIKAL100_tWorkQueue *)params, nNIKAL100_tCreateThreadParamBlock, workQueue);
#endif
   pid_t pid;

   
   pid = kernel_thread( nNIKAL100_threadMain, _paramBlock, SIGCHLD);

   
   _paramBlock->childThreadID = pid;

   
   if (atomic_dec_and_test(&(_paramBlock->counter)))
   {
      nNIKAL100_releaseSingleUseEvent( _paramBlock->threadCreatedEvent );
      _paramBlock->threadCreatedEvent = NULL;
   }
}
#endif
nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_createSystemThread( void *threadContext, nNIKAL100_tI32 (*nNIKAL100_cc threadMain)(void *), nNIKAL100_tUPtr *childPID )
{
   nNIKAL100_tCreateThreadParamBlock paramBlock;
#ifdef nNIKAL150_kKthread
   nLinux_task *task;
#else
   nNIKAL100_tU32 scheduleStatus;
#endif

   
   nNIKAL100_tSingleUseEvent eventNode;
   nNIKAL100_initializeSingleUseEvent(&eventNode);

   paramBlock.threadCreatedEvent = &eventNode;
   paramBlock.threadContext = threadContext;
   paramBlock.threadFunction = threadMain;

#ifdef nNIKAL150_kKthread
   atomic_set(&(paramBlock.counter),1);
   task = kthread_run(nNIKAL100_threadMain, &paramBlock, "niDriverThread");
   if(IS_ERR(task))
   {
      *childPID=0;
      return nNIKAL100_convertLinuxToKALStatus(PTR_ERR(task));
   }
   paramBlock.childThreadID = task->pid;
#else
   atomic_set(&(paramBlock.counter),2);
#ifndef nNIKAL160_kWorkqueueNonDelay
   nNIKAL100_mInitWorkQueue( &paramBlock.workQueue, nNIKAL100_launchThread, &paramBlock );
#else
   nNIKAL100_mInitWorkQueue( &paramBlock.workQueue, (tFuncWithWorkQueuePtrParam)&nNIKAL100_launchThread );
#endif
   scheduleStatus = nNIKAL100_mScheduleWorkQueue( &paramBlock.workQueue );

   if( scheduleStatus == 0 )
   {
      *childPID = 0;
      return nNIKAL100_kStatusThreadCouldNotRun;
   }
#endif

   
   nNIKAL100_waitForSingleUseEvent( &eventNode );

   
   while (paramBlock.threadCreatedEvent)
   {
      nNIKAL100_sleepTimeoutInterruptible(1);
   }

   *childPID = paramBlock.childThreadID;
   return nNIKAL100_kStatusSuccess;
}


nNIKAL100_cc void * nNIKAL150_createMutex( void )
{
#ifdef nNIKAL150_kMutexMethod
   nLinux_mutex *_mutex;  

   if((_mutex = nNIKAL100_malloc( sizeof( nLinux_mutex ) )) != NULL )
   {
      
      mutex_init( _mutex );
   }

   
   return (void *)_mutex;
#else
   return nNIKAL100_createSemaphore(1);
#endif
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL150_destroyMutex( void * mutex )
{
#ifdef nNIKAL150_kMutexMethod
   
   nLinux_mutex *_mutex;
   if(!mutex) return nNIKAL100_kStatusSuccess;
   _mutex = (nLinux_mutex *)mutex;

   
   nNIKAL100_free( _mutex );

   return nNIKAL100_kStatusSuccess;
#else
   return nNIKAL100_destroySemaphore(mutex);
#endif
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL150_acquireMutex( void * mutex )
{
#ifdef nNIKAL150_kMutexMethod
   
   nLinux_mutex *_mutex = (nLinux_mutex *)mutex;

   nNIKAL120_mCheckStackUsage;

   
   mutex_lock( _mutex );

   
   return nNIKAL100_kStatusSuccess;
#else
   return nNIKAL100_acquireSemaphore(mutex);
#endif
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL150_acquireMutexInterruptible( void * mutex )
{
#ifdef nNIKAL150_kMutexMethod
   int status;

   
   nLinux_mutex *_mutex = (nLinux_mutex *)mutex;

   nNIKAL120_mCheckStackUsage;

   status = mutex_lock_interruptible( _mutex );
   return nNIKAL100_convertLinuxToKALStatus(status);
#else
   return nNIKAL100_acquireSemaphoreInterruptible(mutex);
#endif
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL150_acquireMutexZeroTimeout( void * mutex )
{
#ifdef nNIKAL150_kMutexMethod
   
   nLinux_mutex *_mutex = (nLinux_mutex *)mutex;

   nNIKAL120_mCheckStackUsage;

   
   if ( mutex_trylock( _mutex ) == 1 )
   {
      return nNIKAL100_kStatusSuccess;
   }
   else
   {
      return nNIKAL100_kStatusSyncTimedOut;
   }
#else
   return nNIKAL100_acquireSemaphoreZeroTimeout(mutex);
#endif
}


nNIKAL100_cc void nNIKAL150_releaseMutex( void * mutex )
{
#ifdef nNIKAL150_kMutexMethod
   
   nLinux_mutex *_mutex = (nLinux_mutex *)mutex;

   nNIKAL120_mCheckStackUsage;

   
   mutex_unlock( _mutex );

#else
   nNIKAL100_releaseSemaphore(mutex);
#endif
}


nNIKAL100_cc void * nNIKAL100_createSemaphore( nNIKAL100_tI32 initVal )
{
   nLinux_semaphore *_sema;  

   if((_sema = nNIKAL100_malloc( sizeof( nLinux_semaphore ) )) != NULL )
   {
      
      sema_init( _sema, initVal );
   }

   
   return (void *)_sema;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_destroySemaphore( void * sema )
{
   
   nLinux_semaphore *_sema;
   if(!sema) return nNIKAL100_kStatusSuccess;
   _sema = (nLinux_semaphore *)sema;

   
   nNIKAL100_free( _sema );

   return nNIKAL100_kStatusSuccess;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_acquireSemaphore( void * sema )
{
   
   nLinux_semaphore *_sema = (nLinux_semaphore *)sema;

   nNIKAL120_mCheckStackUsage;

   
   down( _sema );

   
   return nNIKAL100_kStatusSuccess;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_acquireSemaphoreInterruptible( void * sema )
{
   int status;

   
   nLinux_semaphore *_sema = (nLinux_semaphore *)sema;

   nNIKAL120_mCheckStackUsage;

   status = down_interruptible( _sema );
   return nNIKAL100_convertLinuxToKALStatus(status);
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_acquireSemaphoreZeroTimeout( void * sema )
{
   
   nLinux_semaphore *_sema = (nLinux_semaphore *)sema;

   nNIKAL120_mCheckStackUsage;

   
   if ( down_trylock( _sema ) == 0 )
   {
      return nNIKAL100_kStatusSuccess;
   }
   else
   {
      return nNIKAL100_kStatusSyncTimedOut;
   }

}


nNIKAL100_cc nNIKAL100_tI32 nNIKAL100_releaseSemaphore( void * sema )
{
   
   nLinux_semaphore *_sema = (nLinux_semaphore *)sema;

   nNIKAL120_mCheckStackUsage;

   
   up( _sema );

   
   return 0;
}



nNIKAL100_cc void * nNIKAL100_createSpinLock( void )
{
   nLinux_spinlock *_spinlock;  

   if(( _spinlock = nNIKAL100_malloc( sizeof( nLinux_spinlock ) )) != NULL )
   {
      
      spin_lock_init( _spinlock );
   }

   
   return (void *)_spinlock;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_destroySpinLock( void *spinlock )
{
   nLinux_spinlock *_spinlock;
   if(!spinlock) return nNIKAL100_kStatusSuccess;
   _spinlock = (nLinux_spinlock *)spinlock;  

   
   if( spin_is_locked( _spinlock ) )
   {
      
      return nNIKAL100_kStatusResourceBusy;
   }

   nNIKAL100_free( _spinlock );

   
   return nNIKAL100_kStatusSuccess;
}


nNIKAL100_cc nNIKAL100_tUPtr nNIKAL110_acquireSpinLockInterrupt( void *spinlock )
{
   nLinux_spinlock *_spinlock = (nLinux_spinlock *)spinlock;  
   nNIKAL100_tUPtr _flags;                              

   spin_lock_irqsave( _spinlock, _flags );

   
   return _flags;
}


nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_acquireSpinLockInterrupt( void *spinlock )
{
   return (nNIKAL100_tU32)nNIKAL110_acquireSpinLockInterrupt( spinlock );
}


nNIKAL100_cc void nNIKAL110_releaseSpinLockInterrupt( void *spinlock, nNIKAL100_tUPtr flags )
{
   
   spin_unlock_irqrestore( (nLinux_spinlock *)spinlock, flags );
}


nNIKAL100_cc void nNIKAL100_releaseSpinLockInterrupt( void *spinlock, nNIKAL100_tU32 flags )
{
   nNIKAL110_releaseSpinLockInterrupt( spinlock, (nNIKAL100_tUPtr)flags );
}


nNIKAL100_cc void nNIKAL100_acquireSpinLockDPC( void * spinlock )
{
   
   spin_lock_bh( (nLinux_spinlock *)spinlock );
}


nNIKAL100_cc void nNIKAL100_releaseSpinLockDPC( void *spinlock )
{
   
   spin_unlock_bh( (nLinux_spinlock *)spinlock );
}


nNIKAL100_cc void nNIKAL100_initializeSingleUseEvent( void *event )
{
   nNIKAL100_tSingleUseEvent *_event = (nNIKAL100_tSingleUseEvent *)event;
   _event->done = 0;
   init_waitqueue_head(&(_event->waitQueue));
}


nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_getSingleUseEventSize( void )
{
   return sizeof( nNIKAL100_tSingleUseEvent );
}


nNIKAL100_cc void nNIKAL100_releaseSingleUseEvent( void *event )
{
   nNIKAL100_tSingleUseEvent *_event = (nNIKAL100_tSingleUseEvent *)event;

   nNIKAL120_mCheckStackUsage;

   _event->done = 1;
   wake_up(&(_event->waitQueue));
}


static void nNIKAL100_timeoutCallback( unsigned long package )
{
   nNIKAL100_tTimerCallbackSpec *_timerCallbackPkg =
      (nNIKAL100_tTimerCallbackSpec *)package;

   _timerCallbackPkg->func( _timerCallbackPkg->context );
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_waitForSingleUseEvent( void *event )
{
   nNIKAL100_tSingleUseEvent *_event = (nNIKAL100_tSingleUseEvent *)event;

   
   nLinux_task *currentTask = current;
   if ((currentTask->mm == NULL) && ((currentTask->flags & PF_EXITING) == 0))
   {
      return nNIKAL100_waitForSingleUseEventInterruptible(event);
   }
   else
   {
      

      
      
#if nNIKAL100_mIsLinux24
      wait_event((_event->waitQueue), _event->done);
#else
      
      DEFINE_WAIT(wait);
      if (!(_event->done))
      {
         prepare_to_wait_exclusive(&(_event->waitQueue), &wait, TASK_UNINTERRUPTIBLE);
         if (!(_event->done))
            schedule();
         finish_wait(&(_event->waitQueue), &wait);
      }
#endif
   }

   nNIKAL120_mCheckStackUsage;

   return nNIKAL100_kStatusSuccess;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_waitForSingleUseEventInterruptible ( void *event )
{
   nNIKAL100_tSingleUseEvent *_event = (nNIKAL100_tSingleUseEvent *)event;
   nNIKAL100_tStatus retVal = nNIKAL100_kStatusSuccess;

#if nNIKAL100_mIsLinux24
   retVal = wait_event_interruptible((_event->waitQueue), _event->done);
#else
   
   DEFINE_WAIT(wait);
   if (!(_event->done))
   {
      prepare_to_wait_exclusive(&(_event->waitQueue), &wait, TASK_INTERRUPTIBLE);
      while (!(_event->done))
      {
         if (signal_pending(current))
         {
            retVal = nNIKAL100_kStatusWaitInterrupted;
            break;
         }
         schedule();
      }
      finish_wait(&(_event->waitQueue), &wait);
   }
#endif

   nNIKAL120_mCheckStackUsage;

   return retVal;
}


static unsigned long calculateTimerExpire( nNIKAL100_tU32 timeout )
{
   unsigned long expires = 0;

   

   
   if( timeout > 2000000 )
   {
      
      expires = ((timeout / 1000) * HZ) + (((timeout % 1000) * HZ) / 1000) + jiffies + 1;
   }
   else
   {
      expires  = ((timeout * HZ) / 1000 ) + jiffies + 1;
   }

   return expires;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_waitForSingleUseEventTimeout
(
   void *event,
   nNIKAL100_tU32 timeout,
   nNIKAL100_tTimerCallbackSpec *timerCallbackPkg
)
{
   nLinux_timerList timer;
   nNIKAL100_tStatus retVal = nNIKAL100_kStatusSuccess;

   nNIKAL120_mCheckStackUsage;

   
   init_timer( &timer );

   timer.expires  = calculateTimerExpire( timeout );
   nNIKAL100_compileTimeAssert(sizeof(nNIKAL100_tTimerCallbackSpec*)==sizeof(unsigned long), "Oops, resizing nNIKAL100_tTimerCallbackSpec* into unsigned long!");
   timer.data     = (unsigned long) timerCallbackPkg;
   timer.function = nNIKAL100_timeoutCallback;

   
   add_timer( &timer );

   
   retVal = nNIKAL100_waitForSingleUseEvent( event );

   del_timer_sync( &timer );

   return retVal;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_waitForSingleUseEventTimeoutInterruptible
(
   void *event,
   nNIKAL100_tU32 timeout,
   nNIKAL100_tTimerCallbackSpec *timerCallbackPkg
)
{
   nLinux_timerList timer;
   nNIKAL100_tStatus retVal = nNIKAL100_kStatusSuccess;

   nNIKAL120_mCheckStackUsage;

   
   init_timer( &timer );

   timer.expires  = calculateTimerExpire( timeout );
   timer.data     = (unsigned long) timerCallbackPkg;
   timer.function = nNIKAL100_timeoutCallback;

   
   add_timer( &timer );

   retVal = nNIKAL100_waitForSingleUseEventInterruptible(event);

   del_timer_sync( &timer );

   return retVal;  
}


static inline nNIKAL100_tPCIDeviceIDTableFooter *nNIKAL100_getPCIDeviceIDTableFooter(
   const nLinux_pciDeviceID *entry)
{
   nNIKAL100_tPCIDeviceIDTableFooter *tableFooter;
   nNIKAL100_tU32 i = 0;

   
   for (; (entry->vendor != 0) || (entry->device != 0) ||
      (entry->subvendor != 0) || (entry->subdevice != 0) || (entry->class != 0); ++entry)
   {}

   
   ++i;
   ++entry;
   tableFooter = (nNIKAL100_tPCIDeviceIDTableFooter*)entry;
   while (tableFooter->signature != nNIKAL100_kPCIDeviceIDTableFooterSignature)
   {
      ++i;
      ++entry;
      tableFooter = (nNIKAL100_tPCIDeviceIDTableFooter*)entry;
   }
   
   return tableFooter;
}


static int nNIKAL100_pciProbe(nLinux_pciDevice *dev, const nLinux_pciDeviceID *id)
{
   nNIKAL100_tDeviceInfo pciDevice;
   int status;
   nNIKAL100_tU8 rev, irqPin;
   nNIKAL100_tU32 i, ioCount;
   nNIKAL100_tPCIDeviceIDTableFooter* tableFooter =
      nNIKAL100_getPCIDeviceIDTableFooter(id);
   nNIKAL100_tDevicePrivateData *devicePrivateData;

   
   devicePrivateData = (nNIKAL100_tDevicePrivateData*)
      nNIKAL100_malloc(sizeof(nNIKAL100_tDevicePrivateData));
   if (devicePrivateData == NULL)
      return -ENOMEM;

   devicePrivateData->driver = tableFooter->driver;

   memset(&pciDevice, 0, sizeof(nNIKAL100_tDeviceInfo));
   pciDevice.osDevice = (void*)dev;
   pciDevice.busType = nNIKAL100_kBusTypePCI;
   pciDevice.version = 1;

   pciDevice.pci.deviceID.bus = dev->bus->number;
   pciDevice.pci.deviceID.slot = PCI_SLOT(dev->devfn);
   pciDevice.pci.deviceID.function = PCI_FUNC(dev->devfn);
   pciDevice.pci.deviceID.vendor = dev->vendor;
   pciDevice.pci.deviceID.device = dev->device;
   pciDevice.pci.deviceID.subSystemVendor = dev->subsystem_vendor;
   pciDevice.pci.deviceID.subSystemDevice = dev->subsystem_device;

   status = pci_enable_device(dev);
   if (status < 0)
   {
      nNIKAL100_free(devicePrivateData);
      return status;
   }

   pci_read_config_byte(dev, PCI_CLASS_REVISION, &rev);
   pciDevice.pci.revision = rev;

   pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &irqPin);
   if (irqPin)
   {
      pciDevice.pci.irqSize = 1;
      pciDevice.pci.irq = dev->irq;
   }

   
   pci_set_master(dev);
   for (i=ioCount=0; i<nNIKAL100_kPCIBARMax; ++i)
   {
      if (pci_resource_len(dev, i) == 0)
         break;
      pciDevice.pci.ioResource[i].start = pci_resource_start(dev, i);
      pciDevice.pci.ioResource[i].size = pci_resource_len(dev, i);
      if (pci_resource_flags(dev, i) & IORESOURCE_IO)
         pciDevice.pci.ioResource[i].flags |= nNIKAL100_kPCIIOResourceFlagTypeIO;
      else if (pci_resource_flags(dev, i) & IORESOURCE_MEM)
         pciDevice.pci.ioResource[i].flags |= nNIKAL100_kPCIIOResourceFlagTypeMem;
      ++ioCount;
   }
   pciDevice.pci.ioResourceSize = ioCount;

   status = devicePrivateData->driver->addDevice(&pciDevice,
      (void*)(id->driver_data), &(devicePrivateData->deviceExtension));
   if (status >= 0)
   {
      pci_set_drvdata(dev, devicePrivateData);
   }
   else
   {
      nNIKAL100_free(devicePrivateData);
      pci_disable_device(dev);
   }

   return status;
}


static void nNIKAL100_pciRemove(nLinux_pciDevice *dev)
{
   nNIKAL100_tDevicePrivateData *devicePrivateData =
      (nNIKAL100_tDevicePrivateData*)pci_get_drvdata(dev);
   devicePrivateData->driver->removeDevice(devicePrivateData->deviceExtension);
   pci_set_drvdata(dev, NULL);
   pci_disable_device(dev);
   nNIKAL100_free(devicePrivateData);
}


nNIKAL100_cc void *nNIKAL100_registerPCIDriver(nNIKAL100_tDriver *driver,
   const nNIKAL100_tText *name, void *pciDevIdTable)
{
   nNIKAL100_tPCIDeviceIDTableFooter* tableFooter =
      (nNIKAL100_tPCIDeviceIDTableFooter*)pciDevIdTable;
   nLinux_pciDriver *pciDriver;
   int ret;

   tableFooter->driver = driver;

   pciDriver = nNIKAL100_malloc(sizeof(nLinux_pciDriver));
   if (pciDriver == NULL)
      return NULL;

   memset(pciDriver, 0, sizeof(nLinux_pciDriver));

   pciDriver->name = (char*)name;
   pciDriver->id_table = tableFooter->table;
   pciDriver->probe = nNIKAL100_pciProbe;
   pciDriver->remove = nNIKAL100_pciRemove;
   pciDriver->suspend = NULL;
   pciDriver->resume = NULL;
   ret = pci_register_driver(pciDriver);
   if ( ret < 0 )
      return NULL;

   return pciDriver;
}


nNIKAL100_cc void nNIKAL100_unregisterPCIDriver(void *pciDriverData, const nNIKAL100_tText ** name,
   void **pciDevIdTable)
{
   nLinux_pciDriver *pciDriver = (nLinux_pciDriver*)pciDriverData;
   nNIKAL100_tPCIDeviceIDTableFooter* tableFooter =
      nNIKAL100_getPCIDeviceIDTableFooter(pciDriver->id_table);
   *name = pciDriver->name;
   *pciDevIdTable = (void*)(tableFooter);
   pci_unregister_driver(pciDriver);
   nNIKAL100_free(pciDriver);
}


nNIKAL100_cc void *nNIKAL100_createPCIDeviceIDTable(nNIKAL100_tU32 size)
{
   nNIKAL100_tText *buffer;
   nNIKAL100_tPCIDeviceIDTableFooter *pciDeviceIDTableFooter;

   
   nNIKAL100_tU32 bufferSize = ((size+1)*sizeof(nLinux_pciDeviceID)) + sizeof(nNIKAL100_tPCIDeviceIDTableFooter);
   buffer = (nNIKAL100_tText*)nNIKAL100_malloc(bufferSize);
   if (buffer == NULL)
      return NULL;

   memset(buffer, 0, bufferSize);

   
   pciDeviceIDTableFooter = (nNIKAL100_tPCIDeviceIDTableFooter*)
      (buffer + ((size+1)*sizeof(nLinux_pciDeviceID)));
   pciDeviceIDTableFooter->signature = nNIKAL100_kPCIDeviceIDTableFooterSignature;
   pciDeviceIDTableFooter->capacity = size;
   pciDeviceIDTableFooter->size = 0;
   pciDeviceIDTableFooter->driver = NULL;
   pciDeviceIDTableFooter->table = (nLinux_pciDeviceID*)buffer;
   return (void*)pciDeviceIDTableFooter;
}


nNIKAL100_cc void nNIKAL100_setPCIDeviceIDTableElement(void *pciDevIdTable,
   nNIKAL100_tU32 vendor, nNIKAL100_tU32 device, nNIKAL100_tU32 subsysVendor,
   nNIKAL100_tU32 subsysDevice, nNIKAL100_tU32 devClass,
   nNIKAL100_tU32 devClassMask, void *privateData)
{
   nNIKAL100_tPCIDeviceIDTableFooter* tableFooter =
      (nNIKAL100_tPCIDeviceIDTableFooter*)pciDevIdTable;
   nLinux_pciDeviceID *idTable = tableFooter->table;
   nNIKAL100_tU32 index = tableFooter->size;

   ++(tableFooter->size);
   nNIKAL100_assertFatal(tableFooter->size <= tableFooter->capacity, "nNIKAL100_setPCIDeviceIDTableElement: pciDeviceIDTable size exceeds capacity\n");

   idTable[index].vendor = vendor;
   idTable[index].device = device;

   
   if (subsysVendor == nNIKAL100_kPCIDontCare && subsysDevice == nNIKAL100_kPCIDontCare)
   {
      idTable[index].subvendor = PCI_ANY_ID;
      idTable[index].subdevice = PCI_ANY_ID;
   }
   else if ( subsysVendor != nNIKAL100_kPCIDontCare && subsysDevice != nNIKAL100_kPCIDontCare)
   {
      idTable[index].subvendor = subsysVendor;
      idTable[index].subdevice = subsysDevice;
   }
   else
   {
      nNIKAL100_assertFatal(0,"nNIKAL100_setPCIDeviceIDTableElement: subsysVendor and subsysDevice should be specified together\n");
   }

   idTable[index].class = devClass;
   idTable[index].class_mask = devClassMask;
   nNIKAL100_compileTimeAssert(sizeof(void*)==sizeof(unsigned long), "Oops, resizing void* into unsigned long!");
   idTable[index].driver_data = (unsigned long)privateData;
}


nNIKAL100_cc void nNIKAL100_destroyPCIDeviceIDTable(void *pciDevIdTable)
{
   nNIKAL100_tPCIDeviceIDTableFooter* tableFooter;
   if(!pciDevIdTable) return;
   tableFooter = (nNIKAL100_tPCIDeviceIDTableFooter*)pciDevIdTable;
   nNIKAL100_free(tableFooter->table);
}


nNIKAL100_cc void nNIKAL100_enumeratePCIBuses(nNIKAL100_tPCIBridgeVisitor visitor,
   void *visitorArgument,
   void *listToEnumerate)
{
   nLinux_listHead *busList = (nLinux_listHead*)listToEnumerate;
   nLinux_listHead *iterator;
   nLinux_pciBus *bus;
   nNIKAL100_tPCIBridgeInfo pciBusInformation;

   
   if (busList == NULL)
      busList = &pci_root_buses;

   
   list_for_each(iterator, busList)
   {
      bus = list_entry(iterator, nLinux_pciBus, node);
      if (bus->self)
      {
         pciBusInformation.deviceID.bus = bus->primary;
         pciBusInformation.deviceID.slot = PCI_SLOT(bus->self->devfn);
         pciBusInformation.deviceID.function = PCI_FUNC(bus->self->devfn);
         pciBusInformation.deviceID.vendor = bus->self->vendor;
         pciBusInformation.deviceID.device = bus->self->device;
         pciBusInformation.deviceID.subSystemVendor = bus->self->subsystem_vendor;
         pciBusInformation.deviceID.subSystemDevice = bus->self->subsystem_device;
         if (bus->self->hdr_type == PCI_HEADER_TYPE_BRIDGE)
         {
            nNIKAL100_tI32 subSystemCapOffset = pci_find_capability(bus->self, PCI_CAP_ID_SSVID);
            if( subSystemCapOffset )
            {
               nNIKAL100_tU32 subSystemID = 0;
               nNIKAL100_pciConfigRead32(bus->self, subSystemCapOffset + nNIKAL100_kSubsystemIDOffset, &subSystemID);
               pciBusInformation.deviceID.subSystemVendor = subSystemID & nNIKAL100_kSubsystemVendorIDMask;
               pciBusInformation.deviceID.subSystemDevice = subSystemID >> nNIKAL100_kSubsystemDeviceIDShift;
            }
         }
         pciBusInformation.primary = bus->primary;
         pciBusInformation.secondary = bus->secondary;
         (*visitor)(visitorArgument, &pciBusInformation);
      }

      
      nNIKAL100_enumeratePCIBuses(visitor, visitorArgument, &(bus->children));
   }
}


#if nNIKAL100_mUSBIsSupported
static inline nNIKAL100_tUSBDeviceIDTableFooter *nNIKAL100_getUSBDeviceIDTableFooter(
   const nLinux_usbDeviceID *entry)
{
   nNIKAL100_tUSBDeviceIDTableFooter *tableFooter;
   nNIKAL100_tU32 i = 0;

   
   for (; (entry->match_flags != 0) ||
          (entry->idVendor != 0) || (entry->idProduct != 0) || (entry->bcdDevice_lo != 0) || (entry->bcdDevice_hi != 0) ||
          (entry->bDeviceClass != 0) || (entry->bDeviceSubClass != 0) || (entry->bDeviceProtocol != 0) ||
          (entry->bInterfaceClass != 0) || (entry->bInterfaceSubClass != 0) || (entry->bInterfaceProtocol != 0);
          ++entry)
   {}

   
   ++i;
   ++entry;
   tableFooter = (nNIKAL100_tUSBDeviceIDTableFooter*)entry;
   while (tableFooter->signature != nNIKAL100_kUSBDeviceIDTableFooterSignature)
   {
      ++i;
      ++entry;
      tableFooter = (nNIKAL100_tUSBDeviceIDTableFooter*)entry;
   }
   
   return tableFooter;
}


static int nNIKAL100_usbProbe(nLinux_usbInterface *interface, const nLinux_usbDeviceID *id)
{
   nLinux_usbDevice * dev;
   nNIKAL100_tDeviceInfo usbDeviceData;
   int status;
   nNIKAL100_tUSBDeviceIDTableFooter* tableFooter =
      nNIKAL100_getUSBDeviceIDTableFooter(id);
   nNIKAL100_tDevicePrivateData *devicePrivateData;

   
   devicePrivateData = (nNIKAL100_tDevicePrivateData*)
      nNIKAL100_malloc(sizeof(nNIKAL100_tDevicePrivateData));
   if (devicePrivateData == NULL)
      return -ENOMEM;

   devicePrivateData->driver = tableFooter->driver;

   dev = interface_to_usbdev(interface);

   memset(&usbDeviceData, 0, sizeof(nNIKAL100_tDeviceInfo));
   usbDeviceData.osDevice = (void*)dev;
   usbDeviceData.busType = nNIKAL100_kBusTypeUSB;
   usbDeviceData.version = 1;
   usbDeviceData.usb.deviceID.vendor = dev->descriptor.idVendor;
   usbDeviceData.usb.deviceID.product = dev->descriptor.idProduct;
   usbDeviceData.usb.deviceID.revision = dev->descriptor.bcdDevice;
   usbDeviceData.usb.deviceID.deviceClass = dev->descriptor.bDeviceClass;
   usbDeviceData.usb.deviceID.deviceSubClass = dev->descriptor.bDeviceSubClass;
   usbDeviceData.usb.deviceID.deviceProtocol = dev->descriptor.bDeviceProtocol;
#ifdef nNIKAL100_kCurrentAlternateSetting
   usbDeviceData.usb.deviceID.interfaceClass = interface->cur_altsetting->desc.bInterfaceClass;
   usbDeviceData.usb.deviceID.interfaceSubClass = interface->cur_altsetting->desc.bInterfaceSubClass;
   usbDeviceData.usb.deviceID.interfaceProtocol = interface->cur_altsetting->desc.bInterfaceProtocol;
   usbDeviceData.usb.interfaceNumber = interface->cur_altsetting->desc.bInterfaceNumber;
#else
   usbDeviceData.usb.deviceID.interfaceClass = interface->altsetting[interface->act_altsetting].desc.bInterfaceClass;
   usbDeviceData.usb.deviceID.interfaceSubClass = interface->altsetting[interface->act_altsetting].desc.bInterfaceSubClass;
   usbDeviceData.usb.deviceID.interfaceProtocol = interface->altsetting[interface->act_altsetting].desc.bInterfaceProtocol;
   usbDeviceData.usb.interfaceNumber = interface->altsetting[interface->act_altsetting].desc.bInterfaceNumber;
#endif
   usbDeviceData.usb.parentDevice = dev;
   usbDeviceData.usb.deviceAddress = dev->devnum;
  
   switch(dev->speed) {
      case USB_SPEED_LOW:
         usbDeviceData.usb.connectionSpeed = nNIKAL120_kUSBSignallingSpeedLow;
         break;
      case USB_SPEED_FULL:
         usbDeviceData.usb.connectionSpeed = nNIKAL120_kUSBSignallingSpeedFull;
         break;
      case USB_SPEED_HIGH:
         usbDeviceData.usb.connectionSpeed = nNIKAL120_kUSBSignallingSpeedHigh;
         break;
      case USB_SPEED_UNKNOWN:
      default:
         usbDeviceData.usb.connectionSpeed = nNIKAL120_kUSBSignallingSpeedUnknown;
         break;
   }


   status = devicePrivateData->driver->addDevice(&usbDeviceData,
      (void*)(id->driver_info), &(devicePrivateData->deviceExtension));
   if (status >= 0)
   {
      usb_set_intfdata(interface, devicePrivateData);
      usb_get_dev(dev);
#ifdef nNIKAL100_kUSBReferenceCountFunctions
      /** Supposedly we are supposed to call this in order to refcount the device,
       * but no existing drivers do, and the symbol isn't even exported from the
       * kernel for modules to use...
      usb_get_intf(interface);
       **/
#endif
   }
   else
   {
      nNIKAL100_free(devicePrivateData);
   }

   return status;
}


static void nNIKAL100_usbDisconnect(nLinux_usbInterface *interface)
{
   nNIKAL100_tDevicePrivateData *devicePrivateData =
      (nNIKAL100_tDevicePrivateData*)usb_get_intfdata(interface);
   devicePrivateData->driver->removeDevice(devicePrivateData->deviceExtension);
   usb_put_dev(interface_to_usbdev(interface));
#ifdef nNIKAL100_kUSBReferenceCountFunctions
   /** Supposedly we are supposed to call this in order to refcount the device,
    * but no existing drivers do, and the symbol isn't even exported from the
    * kernel for modules to use...
   usb_put_intf(interface);
    **/
#endif
   usb_set_intfdata(interface, NULL);
   nNIKAL100_free(devicePrivateData);
}
#endif


nNIKAL100_cc void *nNIKAL100_registerUSBDriver(nNIKAL100_tDriver *driver,
   const nNIKAL100_tText *name, void *usbDevIdTable)
{
#if nNIKAL100_mUSBIsSupported
   nNIKAL100_tUSBDeviceIDTableFooter* tableFooter =
      (nNIKAL100_tUSBDeviceIDTableFooter*)usbDevIdTable;
   nLinux_usbDriver *usbDriver;

   tableFooter->driver = driver;

   usbDriver = nNIKAL100_malloc(sizeof(nLinux_usbDriver));
   if (usbDriver == NULL)
      return NULL;

   memset(usbDriver, 0, sizeof(nLinux_usbDriver));

   #ifdef nNIKAL130_kUSBOwnerMember
      usbDriver->owner = THIS_MODULE;
   #endif
   usbDriver->name = (char*)name;
   usbDriver->probe = nNIKAL100_usbProbe;
   usbDriver->disconnect = nNIKAL100_usbDisconnect;
   usbDriver->ioctl = NULL;
   usbDriver->suspend = NULL;
   usbDriver->resume = NULL;
   usbDriver->id_table = tableFooter->table;
   usb_register(usbDriver);

   return usbDriver;
#else
   return NULL;
#endif
}


nNIKAL100_cc void nNIKAL100_unregisterUSBDriver(void *usbDriverData, const nNIKAL100_tText ** name,
   void **usbDevIdTable)
{
#if nNIKAL100_mUSBIsSupported
   nLinux_usbDriver *usbDriver = (nLinux_usbDriver*)usbDriverData;
   nNIKAL100_tUSBDeviceIDTableFooter* tableFooter =
      nNIKAL100_getUSBDeviceIDTableFooter(usbDriver->id_table);
   *name = usbDriver->name;
   *usbDevIdTable = (void*)(tableFooter);
   usb_deregister(usbDriver);
   nNIKAL100_free(usbDriver);
#else
   
   nNIKAL100_assertFatal(usbDriverData == NULL, "How did you get a non-NULL usbDriverData?  We don't support it on 2.4 Kernels!\n");
#endif
}


nNIKAL100_cc void *nNIKAL100_createUSBDeviceIDTable(nNIKAL100_tU32 size)
{
#if nNIKAL100_mUSBIsSupported
   nNIKAL100_tText *buffer;
   nNIKAL100_tUSBDeviceIDTableFooter *usbDeviceIDTableFooter;

   
   nNIKAL100_tU32 bufferSize = ((size+1)*sizeof(nLinux_usbDeviceID)) + sizeof(nNIKAL100_tUSBDeviceIDTableFooter);
   buffer = (nNIKAL100_tText*)nNIKAL100_malloc(bufferSize);
   if (buffer == NULL)
      return NULL;

   memset(buffer, 0, bufferSize);

   
   usbDeviceIDTableFooter = (nNIKAL100_tUSBDeviceIDTableFooter*)
      (buffer + ((size+1)*sizeof(nLinux_usbDeviceID)));
   usbDeviceIDTableFooter->signature = nNIKAL100_kUSBDeviceIDTableFooterSignature;
   usbDeviceIDTableFooter->capacity = size;
   usbDeviceIDTableFooter->size = 0;
   usbDeviceIDTableFooter->driver = 0;
   usbDeviceIDTableFooter->table = (nLinux_usbDeviceID*)buffer;
   return (void*)usbDeviceIDTableFooter;
#else
   return NULL;
#endif
}


nNIKAL100_cc void nNIKAL100_setUSBDeviceIDTableElement(void *usbDevIdTable,
   nNIKAL100_tU32 vendor, nNIKAL100_tU32 product,
   void *privateData)
{
#if nNIKAL100_mUSBIsSupported
   nNIKAL100_tUSBDeviceIDTableFooter* tableFooter;
   nLinux_usbDeviceID *idTable;
   nNIKAL100_tU32 index;

   nNIKAL100_assertFatal(usbDevIdTable != NULL, "usbDevIdTable must NOT be NULL!\n");

   tableFooter = (nNIKAL100_tUSBDeviceIDTableFooter*)usbDevIdTable;
   idTable = tableFooter->table;
   index = tableFooter->size;

   ++(tableFooter->size);
   nNIKAL100_assertFatal(tableFooter->size <= tableFooter->capacity, "nNIKAL100_setUSBDeviceIDTableElement: usbDeviceIDTable size exceeds capacity\n");

   if( vendor != 0xFFFF ) {
      idTable[index].idVendor = vendor;
      idTable[index].match_flags |= USB_DEVICE_ID_MATCH_VENDOR;
   }
   if( product != 0xFFFF ) {
      idTable[index].idProduct = product;
      idTable[index].match_flags |= USB_DEVICE_ID_MATCH_PRODUCT;
   }



   nNIKAL100_compileTimeAssert(sizeof(void*)==sizeof(unsigned long), "Oops, resizing void* into unsigned long!");
   idTable[index].driver_info = (unsigned long)privateData;
#else
   
#endif
}


nNIKAL100_cc void nNIKAL100_destroyUSBDeviceIDTable(void *usbDevIdTable)
{
#if nNIKAL100_mUSBIsSupported
   nNIKAL100_tUSBDeviceIDTableFooter* tableFooter;
   if(!usbDevIdTable) return;
   tableFooter = (nNIKAL100_tUSBDeviceIDTableFooter*)usbDevIdTable;
   nNIKAL100_free(tableFooter->table);
#else
   
   nNIKAL100_assertFatal(usbDevIdTable == NULL, "How did you get a non-NULL usbDevIdTable?  We don't support it on 2.4 Kernels!\n");
#endif
}

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbSynchronousControlRequest(void* osDevice, nNIKAL100_tUSBSetupPacket* setupPacket, void* data)
{
#if nNIKAL100_mUSBIsSupported
   nNIKAL100_tStatus returnStatus;
   int pipe;
   if( ((setupPacket->bmRequestType) & USB_DIR_IN) == 0 )
   {
      pipe = usb_sndctrlpipe(osDevice, 0);
   }
   else
   {
      pipe = usb_rcvctrlpipe(osDevice, 0);
   }
   returnStatus = usb_control_msg(
                                    osDevice,
                                    pipe,
                                    setupPacket->bRequest,
                                    setupPacket->bmRequestType,
                                    setupPacket->wValue,
                                    setupPacket->wIndex,
                                    data,
                                    setupPacket->wLength,
                                    nNIKAL100_kUSBControlTimeout 
                                 );
   if(returnStatus < 0)
   {
      nNIKAL100_debugPrintf("usb_control_msg() returned %d\n", (int)returnStatus);
      return nNIKAL100_convertLinuxToKALStatus(returnStatus);
   }
   return nNIKAL100_kStatusSuccess;
#else
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbGetDeviceDescriptor(void* osDevice, void* buffer, nNIKAL100_tU32 size)
{
#if nNIKAL100_mUSBIsSupported
   nLinux_usbDevice* usbDevice = (nLinux_usbDevice*)osDevice;
   nNIKAL100_tU32 bytesToCopy = nNIKAL100_mMin(size, usbDevice->descriptor.bLength);

   memcpy(buffer, &(usbDevice->descriptor), bytesToCopy);
   return nNIKAL100_kStatusSuccess;
#else
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}

#if nNIKAL100_mUSBIsSupported
static nNIKAL100_tStatus nNIKAL100_usbFindConfigurationIndex(nLinux_usbDevice* usbDevice, nNIKAL100_tU8 configurationValue, nNIKAL100_tU8* configurationIndex)
{
   nNIKAL100_tU8 configIndex;

   for(configIndex = 0; configIndex < usbDevice->descriptor.bNumConfigurations; ++configIndex)
   {
      if( usbDevice->config[configIndex].desc.bConfigurationValue == configurationValue )
      {
         *configurationIndex = configIndex;
         return nNIKAL100_kStatusSuccess;
      }
   }
   return nNIKAL100_kStatusBadSelector;
}
#endif

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbGetConfigurationDescriptor(void* osDevice, nNIKAL100_tU8 configurationValue, void* buffer, nNIKAL100_tU32 size)
{
#if nNIKAL100_mUSBIsSupported
   nLinux_usbDevice* usbDevice = (nLinux_usbDevice*)osDevice;
   nNIKAL100_tU8 configIndex;
   nNIKAL100_tU32 bytesToCopy;

   nNIKAL100_tStatus kalStatus = nNIKAL100_usbFindConfigurationIndex(usbDevice, configurationValue, &configIndex);
   if( kalStatus != nNIKAL100_kStatusSuccess )
   {
      return kalStatus;
   }

   bytesToCopy = nNIKAL100_mMin(size, usbDevice->config[configIndex].desc.bLength);

   memcpy(buffer, &(usbDevice->config[configIndex].desc), bytesToCopy);
   return nNIKAL100_kStatusSuccess;
#else
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}

#ifdef nNIKAL100_kUSBInterfaceCacheMember
#define nNIKAL100_kUSBInterfaceMember intf_cache
#else
#define nNIKAL100_kUSBInterfaceMember interface
#endif

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbGetInterfaceDescriptor(void* osDevice, nNIKAL100_tU8 configurationValue,
   nNIKAL100_tU8 interfaceNumber, nNIKAL100_tU8 altSettingNumber, void* buffer, nNIKAL100_tU32 size)
{
#if nNIKAL100_mUSBIsSupported
   nLinux_usbDevice* usbDevice = (nLinux_usbDevice*)osDevice;
   nNIKAL100_tU8 configIndex;

   nNIKAL100_tStatus kalStatus = nNIKAL100_usbFindConfigurationIndex(usbDevice, configurationValue, &configIndex);
   if( kalStatus != nNIKAL100_kStatusSuccess )
   {
      return kalStatus;
   }

   if(
      (interfaceNumber < usbDevice->config[configIndex].desc.bNumInterfaces) &&
      (altSettingNumber < usbDevice->config[configIndex].nNIKAL100_kUSBInterfaceMember[interfaceNumber]->num_altsetting)
     )
   {
      nNIKAL100_tU32 bytesToCopy = nNIKAL100_mMin(size, usbDevice->config[configIndex].nNIKAL100_kUSBInterfaceMember[interfaceNumber]->altsetting[altSettingNumber].desc.bLength);

      memcpy(buffer, &(usbDevice->config[configIndex].nNIKAL100_kUSBInterfaceMember[interfaceNumber]->altsetting[altSettingNumber].desc), bytesToCopy);
      return nNIKAL100_kStatusSuccess;
   }
   return nNIKAL100_kStatusBadSelector;
#else
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbGetEndpointDescriptor(void* osDevice, nNIKAL100_tU8 configurationValue,
   nNIKAL100_tU8 interfaceNumber, nNIKAL100_tU8 altSettingNumber, nNIKAL100_tU8 endpointIndex, void* buffer, nNIKAL100_tU32 size)
{
#if nNIKAL100_mUSBIsSupported
   nLinux_usbDevice* usbDevice = (nLinux_usbDevice*)osDevice;
   nNIKAL100_tU8 configIndex;

   nNIKAL100_tStatus kalStatus = nNIKAL100_usbFindConfigurationIndex(usbDevice, configurationValue, &configIndex);
   if( kalStatus != nNIKAL100_kStatusSuccess )
   {
      return kalStatus;
   }

   if(
      (interfaceNumber < usbDevice->config[configIndex].desc.bNumInterfaces) &&
      (altSettingNumber < usbDevice->config[configIndex].nNIKAL100_kUSBInterfaceMember[interfaceNumber]->num_altsetting)
     )
   {
      nLinux_usbInterfaceDescriptorWrapper* hostInterface = &(usbDevice->config[configIndex].nNIKAL100_kUSBInterfaceMember[interfaceNumber]->altsetting[altSettingNumber]);
      if( endpointIndex < hostInterface->desc.bNumEndpoints )
      {
         nNIKAL100_tU32 bytesToCopy = nNIKAL100_mMin(size, hostInterface->endpoint[endpointIndex].desc.bLength);

         memcpy(buffer, &(hostInterface->endpoint[endpointIndex].desc), bytesToCopy);
         return nNIKAL100_kStatusSuccess;
      }
   }
   return nNIKAL100_kStatusBadSelector;
#else
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbSetConfiguration(void* osDevice, nNIKAL100_tU8 configurationValue)
{
#if nNIKAL100_mUSBIsSupported && defined(nNIKAL100_kUSBSetConfiguration)
   int linuxStatus = usb_set_configuration( osDevice, configurationValue );
   if( linuxStatus != 0 )
   {
      nNIKAL100_debugPrintf("usb_set_configuration returned %d\n", linuxStatus);
   }
   return nNIKAL100_convertLinuxToKALStatus(linuxStatus);
#else
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbGetConfiguration(void* osDevice, nNIKAL100_tU8* configurationValuePtr)
{
#if nNIKAL100_mUSBIsSupported
   int returnStatus;
   nNIKAL100_assertFatal(configurationValuePtr != NULL, "Invalid pointer parameter.\n");
   returnStatus = usb_control_msg(  osDevice,
                                    usb_rcvctrlpipe(osDevice, 0),
                                    USB_REQ_GET_CONFIGURATION,
                                    USB_DIR_IN | USB_TYPE_STANDARD | USB_RECIP_DEVICE,
                                    0,
                                    0,
                                    configurationValuePtr,
                                    1,
                                    nNIKAL100_kUSBControlTimeout 
                                 );
   if(returnStatus < 0)
   {
      nNIKAL100_debugPrintf("usb_get_configuration() returned %d\n", (int)returnStatus);
      return nNIKAL100_convertLinuxToKALStatus(returnStatus);
   }
   return nNIKAL100_kStatusSuccess;
#else
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbSetInterface(void* osDevice, nNIKAL100_tU8 interfaceNumber, nNIKAL100_tU8 alternateSetting)
{
#if nNIKAL100_mUSBIsSupported
   int linuxStatus = usb_set_interface( osDevice, interfaceNumber, alternateSetting );
   if( linuxStatus != 0 )
   {
      nNIKAL100_debugPrintf("usb_set_interface returned %d\n", linuxStatus);
   }
   return nNIKAL100_convertLinuxToKALStatus(linuxStatus);
#else
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbGetInterface(void* osDevice, nNIKAL100_tU8 interfaceNumber, nNIKAL100_tU8* alternateSettingPtr)
{
#if nNIKAL100_mUSBIsSupported
   int returnStatus;
   nNIKAL100_assertFatal(alternateSettingPtr != NULL, "Invalid pointer parameter.\n");
   returnStatus = usb_control_msg(  osDevice,
                                    usb_rcvctrlpipe(osDevice, 0),
                                    USB_REQ_GET_INTERFACE,
                                    USB_DIR_IN | USB_TYPE_STANDARD | USB_RECIP_INTERFACE,
                                    0,
                                    interfaceNumber,
                                    alternateSettingPtr,
                                    1,
                                    nNIKAL100_kUSBControlTimeout 
                                 );
   if(returnStatus < 0)
   {
      nNIKAL100_debugPrintf("usb_get_interface() returned %d\n", (int)returnStatus);
      return nNIKAL100_convertLinuxToKALStatus(returnStatus);
   }
   return nNIKAL100_kStatusSuccess;
#else
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}



nNIKAL100_cc nNIKAL100_tURB* nNIKAL100_usbAllocateURB(nNIKAL100_tU32 numISODescriptors)
{
#if nNIKAL100_mUSBIsSupported
   nNIKAL100_tURB* kalURB = (nNIKAL100_tURB*) nNIKAL100_malloc(sizeof(nNIKAL100_tURB) +
                                                               numISODescriptors*sizeof(nNIKAL100_tUSBIsochronousPacketDescriptor));
   if( kalURB != NULL )
   {
      kalURB->osURB = usb_alloc_urb(numISODescriptors, GFP_KERNEL);
      if( kalURB->osURB != NULL )
      {
         if( numISODescriptors > 0 )
         {
            kalURB->isochronous.packetArray = (nNIKAL100_tUSBIsochronousPacketDescriptor*)((void*)kalURB + sizeof(nNIKAL100_tURB));
         }
         else
         {
            kalURB->isochronous.packetArray = NULL;
         }

#ifndef nNIKAL160_kWorkqueueNonDelay
         kalURB->control.osWorkItem = nNIKAL100_malloc(sizeof(nNIKAL100_tWorkQueue));
#else
         kalURB->control.osWorkItem = nNIKAL100_malloc(sizeof(nNIKAL160_tWorkItemWithDataPointer));
#endif
         if( kalURB->control.osWorkItem != NULL )
         {
            /** nNIKAL100_debugPrintf("--> nNIKAL100_usbAllocateURB: Allocated URB 0x%p\n", kalURB->osURB); **/
#ifdef nNIKAL160_kWorkqueueNonDelay
            nNIKAL160_tWorkItemWithDataPointer* tmp_ptr = (nNIKAL160_tWorkItemWithDataPointer*)kalURB->control.osWorkItem;
            tmp_ptr->data = (void *)kalURB;
#endif
            return kalURB; 
         }

         
         usb_free_urb((nLinux_urb*)kalURB->osURB);
      }
      nNIKAL100_free(kalURB);
   }
   return NULL;
#else
   return NULL;
#endif
}



nNIKAL100_cc void nNIKAL100_usbFreeURB(nNIKAL100_tURB* kalURB)
{
#if nNIKAL100_mUSBIsSupported
   if( kalURB == NULL ) return;
   /** nNIKAL100_debugPrintf("<-- nNIKAL100_usbFreeURB: Freeing URB 0x%p\n", kalURB->osURB); **/
   kfree(kalURB->control.osWorkItem);
   usb_free_urb((nLinux_urb*)kalURB->osURB);
   nNIKAL100_free(kalURB);
#else
   
   nNIKAL100_assertFatal(kalURB == NULL, "How did you get a non-NULL kalURB?  We don't support it on 2.4 Kernels!\n");
#endif
}



#if nNIKAL100_mUSBIsSupported
#ifdef nNIKAL160_kIRQRegs
static void nNIKAL100_usbAsynchronousURBCallback(struct urb* osURB, struct pt_regs* unused)
#else
static void nNIKAL100_usbAsynchronousURBCallback(struct urb* osURB)
#endif
{
   nNIKAL100_tURB* kalURB;

   nNIKAL100_assertFatal(osURB != NULL, "Got a callback for a NULL osURB!\n");

   kalURB = (nNIKAL100_tURB*)osURB->context;

   nNIKAL100_assertFatal(kalURB != NULL, "Got a callback for a NULL kalURB!\n");
   nNIKAL100_assertFatal(kalURB->callback != NULL, "Got a callback for a NULL kalURB->callback!\n");

   kalURB->ioStatus = nNIKAL100_convertLinuxToKALStatus(osURB->status);
   kalURB->transferCount = osURB->actual_length;

   /** nNIKAL100_debugPrintf("<----- nNIKAL100_usbAsynchronousURBCallback: URB 0x%p is complete\n", osURB); **/
   kalURB->callback(kalURB);
}

#ifdef nNIKAL160_kIRQRegs
static void nNIKAL100_usbIsochronousURBCallback(struct urb* osURB, struct pt_regs* unused)
#else
static void nNIKAL100_usbIsochronousURBCallback(struct urb* osURB)
#endif
{
   nNIKAL100_tURB* kalURB;
   nNIKAL100_tU32 packetIndex;

   nNIKAL100_assertFatal(osURB != NULL, "Got a callback for a NULL osURB!\n");

   kalURB = (nNIKAL100_tURB*)osURB->context;

   nNIKAL100_assertFatal(kalURB != NULL, "Got a callback for a NULL kalURB!\n");
   nNIKAL100_assertFatal(kalURB->callback != NULL, "Got a callback for a NULL kalURB->callback!\n");

   kalURB = (nNIKAL100_tURB*)osURB->context;
   kalURB->ioStatus = nNIKAL100_convertLinuxToKALStatus(osURB->status);
   kalURB->transferCount = osURB->actual_length;
   kalURB->isochronous.startFrame = osURB->start_frame;
   kalURB->isochronous.numberOfErrors = osURB->error_count;

   for ( packetIndex = 0; packetIndex < kalURB->isochronous.numberOfPackets; ++packetIndex) {
      kalURB->isochronous.packetArray[packetIndex].ioStatus = osURB->iso_frame_desc[packetIndex].status;
      kalURB->isochronous.packetArray[packetIndex].packetTransferCount = osURB->iso_frame_desc[packetIndex].actual_length;
   }
   /** nNIKAL100_debugPrintf("<----- nNIKAL100_usbIsochronousURBCallback: URB 0x%p is complete\n", osURB); **/
   kalURB->callback(kalURB);
}

static void nNIKAL100_buildControlURB(nLinux_usbDevice* usbDevice, nNIKAL100_tURB* kalURB, nLinux_urb* osURB)
{
   unsigned int pipe;

   
   if( (kalURB->transferFlags & nNIKAL100_kUSBTransferFlagDirection) == 0)
   {
      pipe = usb_sndctrlpipe(usbDevice,kalURB->endpoint & USB_ENDPOINT_NUMBER_MASK);
   }
   else
   {
      pipe = usb_rcvctrlpipe(usbDevice,kalURB->endpoint & USB_ENDPOINT_NUMBER_MASK);
   }

   usb_fill_control_urb(
      osURB,
      usbDevice,
      pipe,
      (unsigned char*)&kalURB->control.setupPacket,
      kalURB->buffer,
      kalURB->bufferSize,
      nNIKAL100_usbAsynchronousURBCallback,
      kalURB
      );
}

static void nNIKAL100_buildBulkURB(nLinux_usbDevice* usbDevice, nNIKAL100_tURB* kalURB, nLinux_urb* osURB)
{
   unsigned int pipe;

   
   if( (kalURB->transferFlags & nNIKAL100_kUSBTransferFlagDirection) == 0)
   {
      pipe = usb_sndbulkpipe(usbDevice,kalURB->endpoint & USB_ENDPOINT_NUMBER_MASK);
   }
   else
   {
      pipe = usb_rcvbulkpipe(usbDevice,kalURB->endpoint & USB_ENDPOINT_NUMBER_MASK);
   }

   usb_fill_bulk_urb(
      osURB,
      usbDevice,
      pipe,
      kalURB->buffer,
      kalURB->bufferSize,
      nNIKAL100_usbAsynchronousURBCallback,
      kalURB
      );
}


nLinux_usbEndpointDescriptor* nNIKAL100_usbEndpointAddressToEndpointDescriptor(nLinux_usbDevice* usbDevice, nNIKAL100_tU32 endpoint)
{
   nLinux_usbEndpointDescriptor* endpointDesc;

   
#ifdef nNIKAL100_kUSBDeviceEndpointArray
   nLinux_usbEndpointDescriptorWrapper* endpointDescWrapper;

   if( (endpoint & USB_ENDPOINT_DIR_MASK) == 0)
   {
      endpointDescWrapper = usbDevice->ep_out[endpoint & USB_ENDPOINT_NUMBER_MASK];
      nNIKAL100_assertFatal( endpointDescWrapper != NULL, "No matching ep_out found on the specified interface.\n");
      endpointDesc = &(endpointDescWrapper->desc);
   }
   else
   {
      endpointDescWrapper = usbDevice->ep_in[endpoint & USB_ENDPOINT_NUMBER_MASK];
      nNIKAL100_assertFatal( endpointDescWrapper != NULL, "No matching ep_in found on the specified interface.\n");
      endpointDesc = &(endpointDescWrapper->desc);
   }
#else
   endpointDesc = usb_epnum_to_ep_desc(usbDevice, endpoint);
   nNIKAL100_assertFatal( endpointDesc != NULL, "No matching endpoint found on the specified interface.\n");
#endif
   return endpointDesc;
}

static void nNIKAL100_buildInterruptURB(nLinux_usbDevice* usbDevice, nNIKAL100_tURB* kalURB, nLinux_urb* osURB)
{
   unsigned int pipe;

   nLinux_usbEndpointDescriptor* endpointDesc;

   
   if( (kalURB->transferFlags & nNIKAL100_kUSBTransferFlagDirection) == 0)
   {
      pipe = usb_sndintpipe(usbDevice,kalURB->endpoint & USB_ENDPOINT_NUMBER_MASK);
   }
   else
   {
      pipe = usb_rcvintpipe(usbDevice,kalURB->endpoint & USB_ENDPOINT_NUMBER_MASK);
   }

   osURB->dev                    = usbDevice;
   osURB->pipe                   = pipe;
   osURB->transfer_buffer        = kalURB->buffer;
   osURB->transfer_buffer_length = kalURB->bufferSize;
   osURB->complete               = nNIKAL100_usbAsynchronousURBCallback;
   osURB->context                = kalURB;

   endpointDesc = nNIKAL100_usbEndpointAddressToEndpointDescriptor(usbDevice, kalURB->endpoint);

   nNIKAL100_assertFatal(
      USB_ENDPOINT_XFER_INT == ((endpointDesc->bmAttributes) & USB_ENDPOINT_XFERTYPE_MASK),
      "This is not an interrupt endpoint.\n");
   if( usbDevice->speed == USB_SPEED_HIGH )
   {
      osURB->interval = 1 << min(15, endpointDesc->bInterval - 1);
   }
   else
   {
      osURB->interval = endpointDesc->bInterval;
   }

   nNIKAL100_assertFatal( osURB->interval != 0, "Invalid interval for Interrupt transfer.\n");
}

static void nNIKAL100_buildIsochronousURB(nLinux_usbDevice* usbDevice, nNIKAL100_tURB* kalURB, nLinux_urb* osURB)
{
   unsigned int pipe;

   nLinux_usbEndpointDescriptor *endpointDesc;
   nNIKAL100_tU32 packetIndex;

   
   if( (kalURB->transferFlags & nNIKAL100_kUSBTransferFlagDirection) == 0)
   {
      pipe = usb_sndisocpipe(usbDevice,kalURB->endpoint & USB_ENDPOINT_NUMBER_MASK);
   }
   else
   {
      pipe = usb_rcvisocpipe(usbDevice,kalURB->endpoint & USB_ENDPOINT_NUMBER_MASK);
   }

   osURB->dev                    = usbDevice;
   osURB->pipe                   = pipe;
   osURB->transfer_buffer        = kalURB->buffer;
   osURB->transfer_buffer_length = kalURB->bufferSize;
   osURB->complete               = nNIKAL100_usbIsochronousURBCallback;
   osURB->context                = kalURB;

   endpointDesc = nNIKAL100_usbEndpointAddressToEndpointDescriptor(usbDevice, kalURB->endpoint);

   nNIKAL100_assertFatal(
      USB_ENDPOINT_XFER_ISOC == ((endpointDesc->bmAttributes) & USB_ENDPOINT_XFERTYPE_MASK),
      "This is not an isochronous endpoint.\n");
   if( usbDevice->speed == USB_SPEED_HIGH )
   {
      osURB->interval = 1 << min(15, endpointDesc->bInterval - 1);
   }
   else
   {
      osURB->interval = endpointDesc->bInterval;
   }

   nNIKAL100_assertFatal( osURB->interval != 0, "Invalid interval for Isochronous transfer.\n");

   
   if( kalURB->transferFlags & nNIKAL100_kUSBTransferFlagStartISOTransferASAP )
   {
      osURB->transfer_flags = URB_ISO_ASAP;
   }
   else
   {
      osURB->start_frame = kalURB->isochronous.startFrame;
   }

   osURB->number_of_packets = kalURB->isochronous.numberOfPackets;

   for ( packetIndex = 0; packetIndex < kalURB->isochronous.numberOfPackets; ++packetIndex) {
      osURB->iso_frame_desc[packetIndex].offset = kalURB->isochronous.packetArray[packetIndex].packetOffset;
      osURB->iso_frame_desc[packetIndex].length = kalURB->isochronous.packetArray[packetIndex].packetLength;
   }
}

static void nNIKAL100_usbClearHaltSynchronous(void *params)
{
#ifndef nNIKAL160_kWorkqueueNonDelay
   nNIKAL100_tURB *kalURB = (nNIKAL100_tURB*)params;
#else
   nNIKAL160_tWorkItemWithDataPointer* tmp_ptr = (nNIKAL160_tWorkItemWithDataPointer*)params;
   nNIKAL100_tURB *kalURB = (nNIKAL100_tURB *)tmp_ptr->data;
#endif
   nLinux_usbDevice* usbDevice = (nLinux_usbDevice*)kalURB->osDeviceObject;
   nNIKAL100_tU32 endpointToClear;
   nLinux_usbEndpointDescriptor* endpointDesc;
   int pipe = 0;
   int status;

   endpointToClear = kalURB->control.setupPacket.wIndex;
   endpointDesc = nNIKAL100_usbEndpointAddressToEndpointDescriptor(usbDevice, endpointToClear);

   switch ((endpointDesc->bmAttributes) & USB_ENDPOINT_XFERTYPE_MASK)
   {
      case USB_ENDPOINT_XFER_CONTROL:
         if( (endpointToClear & USB_ENDPOINT_DIR_MASK) == 0)
         {
            pipe = usb_sndctrlpipe(usbDevice,endpointToClear & USB_ENDPOINT_NUMBER_MASK);
         }
         else
         {
            pipe = usb_rcvctrlpipe(usbDevice,endpointToClear & USB_ENDPOINT_NUMBER_MASK);
         }
         break;
      case USB_ENDPOINT_XFER_BULK:
         if( (endpointToClear & USB_ENDPOINT_DIR_MASK) == 0)
         {
            pipe = usb_sndbulkpipe(usbDevice,endpointToClear & USB_ENDPOINT_NUMBER_MASK);
         }
         else
         {
            pipe = usb_rcvbulkpipe(usbDevice,endpointToClear & USB_ENDPOINT_NUMBER_MASK);
         }
         break;
      case USB_ENDPOINT_XFER_ISOC:
         if( (endpointToClear & USB_ENDPOINT_DIR_MASK) == 0)
         {
            pipe = usb_sndisocpipe(usbDevice,endpointToClear & USB_ENDPOINT_NUMBER_MASK);
         }
         else
         {
            pipe = usb_rcvisocpipe(usbDevice,endpointToClear & USB_ENDPOINT_NUMBER_MASK);
         }
         break;
      case USB_ENDPOINT_XFER_INT:
         if( (endpointToClear & USB_ENDPOINT_DIR_MASK) == 0)
         {
            pipe = usb_sndintpipe(usbDevice,endpointToClear & USB_ENDPOINT_NUMBER_MASK);
         }
         else
         {
            pipe = usb_rcvintpipe(usbDevice,endpointToClear & USB_ENDPOINT_NUMBER_MASK);
         }
         break;
   }

   status = usb_clear_halt (usbDevice, pipe);

   kalURB->ioStatus = nNIKAL100_convertLinuxToKALStatus(status);
   kalURB->transferCount = 0;
   kalURB->callback(kalURB);
}

static nNIKAL100_tStatus nNIKAL100_usbClearHaltAsync(nNIKAL100_tURB* kalURB)
{
   int isScheduleSuccess;

#ifndef nNIKAL160_kWorkqueueNonDelay
   nNIKAL100_mInitWorkQueue((nNIKAL100_tWorkQueue*)(kalURB->control.osWorkItem),
      nNIKAL100_usbClearHaltSynchronous, kalURB);
#else
   nNIKAL100_mInitWorkQueue((nNIKAL100_tWorkQueue*)(kalURB->control.osWorkItem),
      (tFuncWithWorkQueuePtrParam)&nNIKAL100_usbClearHaltSynchronous);
#endif
   isScheduleSuccess = nNIKAL100_mScheduleWorkQueue(
      (nNIKAL100_tWorkQueue*)(kalURB->control.osWorkItem));

   
   nNIKAL100_assertFatal(isScheduleSuccess, "Failed to schedule a work item that's just initialized\n");

   return nNIKAL100_kStatusSuccess;
}
#endif


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbSubmitURB(nNIKAL100_tURB* kalURB)
{
#if nNIKAL100_mUSBIsSupported
   nLinux_usbDevice* usbDevice;
   nLinux_urb* osURB;
   int returnStatus;

   nNIKAL100_assertFatal(kalURB != NULL, "Trying to submit a NULL kalURB!\n");
   nNIKAL100_assertFatal(kalURB->osURB != NULL, "Trying to submit a NULL osURB!\n");
   nNIKAL100_assertFatal(kalURB->osDeviceObject != NULL, "Trying to submit a URB with a NULL osDeviceObject!\n");
   nNIKAL100_assertFatal(kalURB->callback != NULL, "Trying to submit a URB with a NULL callback!\n");

   usbDevice = (nLinux_usbDevice*)kalURB->osDeviceObject;
   osURB = (nLinux_urb*)kalURB->osURB;

   switch(kalURB->transferType)
   {
      case nNIKAL100_kUSBTransferTypeControl:
      {
         if((kalURB->control.setupPacket.bRequest == USB_REQ_CLEAR_FEATURE) &&
            (kalURB->control.setupPacket.bmRequestType == USB_RECIP_ENDPOINT) &&
            (kalURB->control.setupPacket.wValue == 0 /** ENDPOINT_HALT **/))
         {
            
            
            return nNIKAL100_usbClearHaltAsync(kalURB);
         }
         else
         {
            nNIKAL100_buildControlURB(usbDevice, kalURB, osURB);
         }
         break;
      }
      case nNIKAL100_kUSBTransferTypeBulk:
      {
         nNIKAL100_buildBulkURB(usbDevice, kalURB, osURB);
         break;
      }
      case nNIKAL100_kUSBTransferTypeInterrupt:
      {
         nNIKAL100_buildInterruptURB(usbDevice, kalURB, osURB);
         break;
      }
      case nNIKAL100_kUSBTransferTypeIsochronous:
      {
         nNIKAL100_buildIsochronousURB(usbDevice, kalURB, osURB);
         break;
      }
      default:
         nNIKAL100_assertFatal(0, "Invalid USB transfer type\n");
         break;
   }

   /** nNIKAL100_debugPrintf("-----> nNIKAL100_usbSubmitURB: Submitted URB 0x%p\n", osURB); **/
   #ifdef nNIKAL130_kURBAsyncUnlink
      osURB->transfer_flags |= URB_ASYNC_UNLINK;
   #endif
   osURB->status = 0;
   /** nNIKAL100_debugPrintf("usb_submit_urb(\n   pipe = 0x%8.8x\n   buffer = 0x%p\n   buffer_size = %d\n   interval = %d\n)\n", osURB->pipe, osURB->transfer_buffer, osURB->transfer_buffer_length, osURB->interval); **/
   returnStatus = usb_submit_urb(osURB, GFP_ATOMIC);

   
   if(returnStatus != 0)
   {
      nNIKAL100_debugPrintf("usb_submit_urb() returned %d\n", returnStatus);
      kalURB->ioStatus = nNIKAL100_convertLinuxToKALStatus(returnStatus);
      kalURB->transferCount = 0;
      kalURB->callback(kalURB);
   }

   return nNIKAL100_convertLinuxToKALStatus(returnStatus);
#else
   
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_usbUnlinkURB(nNIKAL100_tURB* kalURB)
{
#if nNIKAL100_mUSBIsSupported
   int unlinkStatus;

   nNIKAL100_assertFatal(kalURB != NULL, "Trying to unlink a NULL kal URB!\n");
   nNIKAL100_assertFatal(kalURB->osURB != NULL, "Trying to unlink a NULL OS URB!\n");
   #ifdef nNIKAL130_kURBAsyncUnlink
      nNIKAL100_assertFatal( ((nLinux_urb*)kalURB->osURB)->transfer_flags & URB_ASYNC_UNLINK, "The URB is going to be unlinked synchronously?\n");
   #endif

   /** nNIKAL100_debugPrintf("++++++ nNIKAL100_usbUnlinkURB: URB 0x%p is being unlinked\n", kalURB->osURB); **/

   unlinkStatus = usb_unlink_urb((nLinux_urb*)kalURB->osURB);
   if( unlinkStatus != -EINPROGRESS )
   {
      nNIKAL100_debugPrintf("nNIKAL100_usbUnlinkURB(): usb_unlink_urb(%p) returned %d\n", (nLinux_urb*)kalURB->osURB, unlinkStatus);
   }
   
   return nNIKAL100_kStatusSuccess;
#else
   
   return nNIKAL100_kStatusFeatureNotSupported;
#endif
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_pciConfigWrite8 (
   const void * osDevObject,
   const nNIKAL100_tU32 address,
   const nNIKAL100_tU8 data)
{
   if (pci_write_config_byte((nLinux_pciDevice *)osDevObject, address, data))
      return nNIKAL100_kStatusOSFault;
   else
      return nNIKAL100_kStatusSuccess;
}

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_pciConfigWrite16 (
   const void * osDevObject,
   const nNIKAL100_tU32 address,
   const nNIKAL100_tU16 data)
{
   if (pci_write_config_word((nLinux_pciDevice *)osDevObject, address, data))
      return nNIKAL100_kStatusOSFault;
   else
      return nNIKAL100_kStatusSuccess;
}

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_pciConfigWrite32 (
   const void * osDevObject,
   const nNIKAL100_tU32 address,
   const nNIKAL100_tU32 data)
{
   if (pci_write_config_dword((nLinux_pciDevice *)osDevObject, address, data))
      return nNIKAL100_kStatusOSFault;
   else
      return nNIKAL100_kStatusSuccess;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_pciConfigRead8 (
   const void * osDevObject,
   const nNIKAL100_tU32 address,
   nNIKAL100_tU8 *data)
{
   if (pci_read_config_byte((nLinux_pciDevice *)osDevObject, address, data))
      return nNIKAL100_kStatusOSFault;
   else
      return nNIKAL100_kStatusSuccess;
}

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_pciConfigRead16 (
   const void * osDevObject,
   const nNIKAL100_tU32 address,
   nNIKAL100_tU16 *data)
{
   if (pci_read_config_word((nLinux_pciDevice *)osDevObject, address, data))
      return nNIKAL100_kStatusOSFault;
   else
      return nNIKAL100_kStatusSuccess;
}

nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_pciConfigRead32 (
   const void * osDevObject,
   const nNIKAL100_tU32 address,
   nNIKAL100_tU32 *data)
{
   if (pci_read_config_dword((nLinux_pciDevice *)osDevObject, address, (u32*)data))
      return nNIKAL100_kStatusOSFault;
   else
      return nNIKAL100_kStatusSuccess;
}


nNIKAL100_cc void nNIKAL100_ioWrite8 (
   const nNIKAL100_tUPtr address,
   const nNIKAL100_tU8 data)
{
   outb(data, address);
}

nNIKAL100_cc void nNIKAL100_ioWrite16 (
   const nNIKAL100_tUPtr address,
   const nNIKAL100_tU16 data)
{
   outw(data, address);
}

nNIKAL100_cc void nNIKAL100_ioWrite32 (
   const nNIKAL100_tUPtr address,
   const nNIKAL100_tU32 data)
{
   outl(data, address);
}


nNIKAL100_cc nNIKAL100_tU8 nNIKAL100_ioRead8 (
   const nNIKAL100_tUPtr address)
{
   return inb(address);
}

nNIKAL100_cc nNIKAL100_tU16 nNIKAL100_ioRead16 (
   const nNIKAL100_tUPtr address)
{
   return inw(address);
}

nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_ioRead32 (
   const nNIKAL100_tUPtr address)
{
   return inl(address);
}



nNIKAL100_cc void nNIKAL100_memoryWrite8 (
   const nNIKAL100_tUPtr address,
   const nNIKAL100_tU8 data)
{
   writeb(data, (void __iomem *)address);
}

nNIKAL100_cc void nNIKAL100_memoryWrite16 (
   const nNIKAL100_tUPtr address,
   const nNIKAL100_tU16 data)
{
   writew(data, (void __iomem *)address);
}

nNIKAL100_cc void nNIKAL100_memoryWrite32 (
   const nNIKAL100_tUPtr address,
   const nNIKAL100_tU32 data)
{
   writel(data, (void __iomem *)address);
}


nNIKAL100_cc nNIKAL100_tU8 nNIKAL100_memoryRead8 (
   const nNIKAL100_tUPtr address)
{
   return readb((void __iomem *)address);
}

nNIKAL100_cc nNIKAL100_tU16 nNIKAL100_memoryRead16 (
   const nNIKAL100_tUPtr address)
{
   return readw((void __iomem *)address);
}

nNIKAL100_cc nNIKAL100_tU32 nNIKAL100_memoryRead32 (
   const nNIKAL100_tUPtr address)
{
   return readl((void __iomem *)address);
}


#if nNIKAL100_mIsLinux24
static void nNIKAL100_dispatchHardwareInterrupt(int irq, void *context,
   nLinux_registers * registers)
{
   nNIKAL100_tInterruptCallbackSpec *interruptCallbackSpec =
      (nNIKAL100_tInterruptCallbackSpec*)context;
   interruptCallbackSpec->func(irq, interruptCallbackSpec->context);
}
#else
#ifdef nNIKAL160_kIRQRegs
static irqreturn_t nNIKAL100_dispatchHardwareInterrupt(int irq, void *context,
   nLinux_registers * registers)
#else
static irqreturn_t nNIKAL100_dispatchHardwareInterrupt(int irq, void *context)
#endif
{
   nNIKAL100_tInterruptCallbackSpec *interruptCallbackSpec =
      (nNIKAL100_tInterruptCallbackSpec*)context;
   if (interruptCallbackSpec->func(irq, interruptCallbackSpec->context))
      return IRQ_HANDLED;
   else
      return IRQ_NONE;
}
#endif


static void nNIKAL100_dispatchDPC(unsigned long interruptData)
{
   nNIKAL100_tDPCCallbackSpec *dpcCallback =
      (nNIKAL100_tDPCCallbackSpec *)interruptData;
   dpcCallback->func(dpcCallback->context);
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_reserveInterrupt (nNIKAL100_tU32 irq, void *context,
   const nNIKAL100_tText *deviceName)
{

#ifndef IRQF_DISABLED
   #define IRQF_DISABLED SA_INTERRUPT
   #define IRQF_SHARED SA_SHIRQ
#endif
   int status = request_irq(irq, nNIKAL100_dispatchHardwareInterrupt,
                            IRQF_DISABLED | IRQF_SHARED, deviceName, context);

   if (status == -ENOSYS)
      return nNIKAL100_kStatusBadDevice;
   else
      return nNIKAL100_convertLinuxToKALStatus(status);
}


nNIKAL100_cc void nNIKAL100_relinquishInterrupt (nNIKAL100_tU32 irq, void *context)
{
   free_irq(irq, context);
}


nNIKAL100_cc void * nNIKAL100_createDPC(nNIKAL100_tDPCCallback dpcCallback)
{
   nNIKAL100_tDPC *dpc =
      (nNIKAL100_tDPC *)nNIKAL100_malloc(sizeof(nNIKAL100_tDPC));
   if (dpc == NULL)
   {
      return NULL;
   }
   tasklet_init(&(dpc->tasklet), nNIKAL100_dispatchDPC, (unsigned long)(&(dpc->callback)));
   dpc->callback.func = dpcCallback;
   return dpc;
}


nNIKAL100_cc void nNIKAL100_destroyDPC (void * dpc)
{
   nNIKAL100_free(dpc);
}


nNIKAL100_cc void nNIKAL100_scheduleDPC (void * dpc, void *context)
{
   nNIKAL100_tDPC *_dpc = (nNIKAL100_tDPC *)dpc;
   _dpc->callback.context = context;
   tasklet_schedule(&(_dpc->tasklet));
}


nNIKAL100_cc void nNIKAL100_unmapPhysicalFromKernel (void __iomem * memoryArea)
{
   iounmap(memoryArea);
}


nNIKAL100_cc void __iomem * nNIKAL100_mapPhysicalToKernel (nNIKAL100_tUPtr physicalAddress,
   nNIKAL100_tUPtr byteCount)
{
   return ioremap(physicalAddress, byteCount);
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_loadDriver(const nNIKAL100_tText* driverName)
{
   int status = request_module("%s", driverName);

   
   if (status == -ENOSYS)
      return nNIKAL100_kStatusFeatureDisabled;
   else
      return nNIKAL100_convertLinuxToKALStatus(status);
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_incrementDriverRefcount(void* driverHandle)
{
#if nNIKAL100_mIsLinux24
   if (try_inc_mod_count((nLinux_driver*)driverHandle))
#else
   if (try_module_get((nLinux_driver*)driverHandle))
#endif
   {
      return nNIKAL100_kStatusSuccess;
   }
   else
   {
      return nNIKAL100_kStatusResourceNotAvailable;
   }
}


nNIKAL100_cc void nNIKAL100_decrementDriverRefcount(void* driverHandle)
{
   nLinux_driver* module = (nLinux_driver*) driverHandle;
#if nNIKAL100_mIsLinux24
   __MOD_DEC_USE_COUNT(module);
#else
   module_put(module);
#endif
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_unmapKernelFromUser (
   nNIKAL100_tUPtr mappedAddr,
   nNIKAL100_tUPtr inSize,
   const void *osData )
{
   void *addr;
   nNIKAL100_tUPtr size, offset;
   nNIKAL100_tStatus status;

   nNIKAL100_compileTimeAssert(sizeof(void*)==sizeof(nNIKAL100_tUPtr), "Oops, resizing void* into nNIKAL100_tUPtr!\n");
   nNIKAL100_pageAlignMemoryBlock((void *)mappedAddr, inSize, &addr, &size, &offset);

   nNIKAL100_compileTimeAssert(sizeof(void*)==sizeof(nNIKAL100_tUPtr), "Oops, resizing void* into nNIKAL100_tUPtr!\n");

   status = nNIKAL100_doMUnmap((nNIKAL100_tUPtr)addr, size);
   

   return status;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_mapKernelToUser (
   void * const file,
   nNIKAL100_tUPtr sourceAddr,
   nNIKAL100_tUPtr inSize,
   nNIKAL100_tMemoryAddressType addrType,
   nNIKAL100_tUPtr *targetAddr,
   void **osData)
{
   void *addr;
   nNIKAL100_tUPtr size, offset, kernelAddress, i, outAddress;
   nNIKAL100_tU32 pageSize;
   const nNIKAL100_tMMapData pageMapData =
   {
      .callingThreadID = current->pid,
      .addressType = addrType
   };

   nNIKAL100_compileTimeAssert(sizeof(void*)==sizeof(nNIKAL100_tUPtr), "Oops, resizing void* into nNIKAL100_tUPtr!\n");
   nNIKAL100_pageAlignMemoryBlock((void *)sourceAddr, inSize, &addr, &size, &offset);

   
   if (addrType == nNIKAL100_kMemoryAddressTypeVirtual)
   {
      nNIKAL100_assertFatal(osData != NULL,
         "nNIKAL100_mapKernelToUser(): when addrType is virtual, osData must NOT be NULL\n");
      *osData = addr;
      if (nNIKAL100_isKernelContiguousPointer(addr))
      {
         pageSize = nNIKAL100_getPageSize();
         kernelAddress = (nNIKAL100_tUPtr)addr;
         for (i = kernelAddress; i < kernelAddress+size; i+=pageSize)
         {
            nNIKAL100_incrementPageRefcount((void*)i);
         }
      }
   }

   down(&nNIKAL100_sMMapSema);
   nNIKAL100_gMMapDataPtr = &pageMapData;

   outAddress = nNIKAL100_doMMap(file,
      0 , size, (nNIKAL100_tUPtr)addr);

   nNIKAL100_gMMapDataPtr = &nNIKAL100_gMMapData;
   up(&nNIKAL100_sMMapSema);

   

   
   if ( ((nNIKAL100_tIPtr)outAddress <= -1) && ((nNIKAL100_tIPtr)outAddress >= -256) )
   {
      *targetAddr = 0;
      if ((addrType == nNIKAL100_kMemoryAddressTypeVirtual))
      {
         *osData = NULL;
         if (nNIKAL100_isKernelContiguousPointer(addr))
         {
            pageSize = nNIKAL100_getPageSize();
            kernelAddress = (nNIKAL100_tUPtr)addr;
            for (i = kernelAddress; i < kernelAddress+size; i+=pageSize)
            {
               nNIKAL100_decrementPageRefcount((void*)i);
            }
         }
      }

      return nNIKAL100_kStatusOSFault;
   }
   *targetAddr = outAddress+offset;
   return nNIKAL100_kStatusSuccess;
}

static void nNIKAL100_setVMANoCopy(void *pageAlignedInPointer,
   nNIKAL100_tUPtr size, nNIKAL100_tMemPageLockToken *pageLockToken, nNIKAL100_tStatus *status)
{
   nLinux_vmArea *vmArea;
   nNIKAL100_tUPtr ptr, inPointer;
   nNIKAL100_tU32 i=0;
   nNIKAL100_tU32 vmaCount = 0;
   nNIKAL100_tVMAAddressPair *vmaAddressPairs = NULL;
   nNIKAL100_tStatus localStatus = nNIKAL100_kStatusSuccess;
   nNIKAL100_tMapIterator mapIterator;

   nNIKAL100_assertFatal(status != NULL, "NULL status passed to internal function\n");
   if (nNIKAL100_statusIsFatal(*status))
      return;

   
   if (current->mm == NULL)
   {
      nNIKAL100_debugPrintf("nNIKAL100_setVMANoCopy(): called when user mode address space is already destroyed\n");
      *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusResourceNotAvailable);
      return;
   }

   
   down_write(&(current->mm->mmap_sem));

   
   inPointer = (nNIKAL100_tUPtr)pageAlignedInPointer;
   ptr = inPointer;
   do
   {
      vmArea = nNIKAL100_getVMA(ptr);
      if (vmArea == NULL)
      {
         nNIKAL100_debugPrintf("nNIKAL100_setVMANoCopy(): inPtr: %p, size: 0x%lx: cannot find VMA at address: %lx\n", pageAlignedInPointer, size, ptr);
         up_write(&(current->mm->mmap_sem));
         *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusBadPointer);
         return;
      }
#ifdef CONFIG_HUGETLB_PAGE
      if (vmArea->vm_flags & VM_HUGETLB)
      {
         nNIKAL100_debugPrintf("nNIKAL100_setVMANoCopy(): inPtr: %p, size: 0x%lx: Found VM_HUGETLB VMA: %p\n", pageAlignedInPointer, size, vmArea);
         up_write(&(current->mm->mmap_sem));
         *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusFeatureNotSupported);
         return;
      }
#endif

      ptr = vmArea->vm_end;
      ++vmaCount;
   } while (ptr < inPointer+size);

   
   pageLockToken->vmaArraySize = vmaCount;
   if (vmaCount > 1)
   {
      vmaAddressPairs = (nNIKAL100_tVMAAddressPair*)nNIKAL100_malloc(vmaCount * sizeof(nNIKAL100_tVMAAddressPair));
      if (vmaAddressPairs == NULL)
      {
         nNIKAL100_debugPrintf("nNIKAL100_setVMANoCopy(): Failed to allocate memory for vmaAddrPairs: vmaCount: %ld, each data sizeof(): %ld\n", (unsigned long)vmaCount, (unsigned long)sizeof(nNIKAL100_tVMAAddressPair));
         up_write(&(current->mm->mmap_sem));
         *status = nNIKAL100_statusSelect(*status, nNIKAL100_kStatusMemoryFull);
         return;
      }
      pageLockToken->vma = vmaAddressPairs;
   }
   else
      pageLockToken->vma = vmArea;

   
   ptr = inPointer;
   for (i=0; i<vmaCount; ++i)
   {
      if (vmaCount > 1)
      {
         vmArea = nNIKAL100_getVMA(ptr);
         vmaAddressPairs[i].vma = vmArea;
         vmaAddressPairs[i].address = ptr;
      }
      else
      {
         
      }

      nNIKAL100_mapFind(&nNIKAL100_sPageLockMap, (nNIKAL100_tUPtr)vmArea, &mapIterator);
      if (mapIterator.keyValuePtr != NULL)
      {
         ++(mapIterator.keyValuePtr->value);
      }
      else
      {
         if (((vmArea->vm_flags & VM_DONTCOPY) == 0) && ((vmArea->vm_flags & (VM_SHARED | VM_MAYWRITE)) == VM_MAYWRITE))
         {
            localStatus = nNIKAL100_mapInsert(&nNIKAL100_sPageLockMap, (nNIKAL100_tUPtr)vmArea, 1);
            if (nNIKAL100_statusIsNonfatal(localStatus))
            {
               vmArea->vm_flags |= VM_DONTCOPY;
               
            }
            else
               break; 
         }
         else
         {
            
            if (vmaAddressPairs)
            {
               vmaAddressPairs[i].vma = NULL;
               vmaAddressPairs[i].address = 0;
            }
            else
               pageLockToken->vma = NULL;
         }
      }

      
      ptr = vmArea->vm_end;
   }

   if (nNIKAL100_statusIsFatal(localStatus))
   {
      
      for (; i>0; --i)
      {
         vmArea = vmaAddressPairs[i-1].vma;
         nNIKAL100_mapFind(&nNIKAL100_sPageLockMap, (nNIKAL100_tUPtr)vmArea, &mapIterator);
         if (mapIterator.keyValuePtr != NULL)
         {
            --(mapIterator.keyValuePtr->value);
            if (mapIterator.keyValuePtr->value == 0)
            {
               nNIKAL100_mapRemove(&nNIKAL100_sPageLockMap, &mapIterator);
               vmArea->vm_flags &= ~VM_DONTCOPY;
               
            }
         }
      }
      nNIKAL100_free(vmaAddressPairs);
      pageLockToken->vmaArraySize = 0;
      pageLockToken->vma = NULL;

      *status = nNIKAL100_statusSelect(*status, localStatus);
   }

   up_write(&(current->mm->mmap_sem));
}

static void nNIKAL100_unsetVMANoCopy(nNIKAL100_tMemPageLockToken *pageLockToken)
{
   nLinux_vmArea *vmArea = NULL;
   nNIKAL100_tUPtr ptr = 0;
   nNIKAL100_tU32 i=0;
   nNIKAL100_tVMAAddressPair *vmaAddressPairs = NULL;
   nNIKAL100_tMapIterator mapIterator;

   
   if (current->mm != NULL)
      down_write(&(current->mm->mmap_sem));

   if (pageLockToken->vmaArraySize == 1)
   {
      vmArea = (nLinux_vmArea*)(pageLockToken->vma);
      ptr = (nNIKAL100_tUPtr)(pageLockToken->originalPointer);
   }
   else
   {
      vmaAddressPairs = (nNIKAL100_tVMAAddressPair*)(pageLockToken->vma);
      nNIKAL100_assertFatal(vmaAddressPairs != NULL, "nNIKAL100_pageUnlockUserPointer(): pageLockToken seems to be corrupted\n");
   }

   for (i=0; i<pageLockToken->vmaArraySize; ++i)
   {
      if (vmaAddressPairs)
      {
         vmArea = vmaAddressPairs[i].vma;
         ptr = vmaAddressPairs[i].address;
      }

      nNIKAL100_mapFind(&nNIKAL100_sPageLockMap, (nNIKAL100_tUPtr)vmArea, &mapIterator);
      if (mapIterator.keyValuePtr != NULL)
      {
         --(mapIterator.keyValuePtr->value);
         if (mapIterator.keyValuePtr->value == 0)
         {
            nNIKAL100_mapRemove(&nNIKAL100_sPageLockMap, &mapIterator);
            if (vmArea == nNIKAL100_getVMA(ptr))
            {
               
               vmArea->vm_flags &= ~VM_DONTCOPY;
            }
            else
            {
               
               
            }
         }
      }
   }

   nNIKAL100_free(vmaAddressPairs);
   pageLockToken->vmaArraySize = 0;
   pageLockToken->vma = NULL;

   if (current->mm != NULL)
      up_write(&(current->mm->mmap_sem));
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_pageLockUserPointer(
   void __user * inPointer,
   const nNIKAL100_tUPtr inSize,
   void ** outPageLockedPointer,
   nNIKAL100_tMemPageLockToken *pageLockToken)
{
   
   nNIKAL100_tU32 pageSize;
   void *pageAlignedInPointer;
   void *reusableIORemapAddress;
   void *tmpJunkPtr;
   nNIKAL100_tUPtr inPointerOffset, size;
   nNIKAL100_tStatus status = nNIKAL100_kStatusSuccess;
   size = inSize;
   *outPageLockedPointer = NULL;

   down(&nNIKAL100_sPageLockOperationSema);
   pageSize = nNIKAL100_getPageSize();

   nNIKAL100_pageAlignMemoryBlock(inPointer, inSize, &pageAlignedInPointer, &size, &inPointerOffset);
   
   reusableIORemapAddress = (void*)(0-pageSize-size);

   if ((*outPageLockedPointer = nNIKAL100_mapPhysicalToKernel(
            (nNIKAL100_tUPtr)(reusableIORemapAddress), size)) != NULL)
   {
      if ((pageLockToken->pageLockDescriptor = nNIKAL100_mapUserKIOBuf(pageAlignedInPointer, size, &status)) != NULL)
      {
         nNIKAL100_setVMANoCopy(pageAlignedInPointer, size, pageLockToken, &status);
         if (nNIKAL100_statusIsNonfatal(status))
         {
            
            nNIKAL100_copyUserPageTable(pageAlignedInPointer, *outPageLockedPointer, size,
               &(pageLockToken->originalPageTableEntry));

            
            if ((tmpJunkPtr = nNIKAL100_mapPhysicalToKernel((nNIKAL100_tUPtr)(reusableIORemapAddress), pageSize)) != NULL)
            {
               nNIKAL100_unmapPhysicalFromKernel(tmpJunkPtr);

               pageLockToken->originalPointer = pageAlignedInPointer;
               pageLockToken->pageLockedPointer = *outPageLockedPointer;
               pageLockToken->bufferSize = size;
               pageLockToken->addressOrigin = nNIKAL100_kAddressOriginUser;
               
               *outPageLockedPointer = (void*)((nNIKAL100_tUPtr)(*outPageLockedPointer)+inPointerOffset);
               up(&nNIKAL100_sPageLockOperationSema);
               return nNIKAL100_kStatusSuccess;
            }
            else
            {
               nNIKAL100_debugPrintf("_pageLockUserPointer: Failed to do an operation to cause TLB flush\n");
               status = nNIKAL100_statusSelect(status, nNIKAL100_kStatusMemoryFull);
            }

            
            nNIKAL100_setKernelPageTable(&(pageLockToken->originalPageTableEntry),
               pageLockToken->pageLockedPointer, pageLockToken->bufferSize);

            nNIKAL100_unsetVMANoCopy(pageLockToken);
         }
         else
         {
            nNIKAL100_debugPrintf("_pageLockUserPointer: Failed to mark VMAs with VM_DONTCOPY (%p, %ld)\n", pageAlignedInPointer, size);
         }

         
         nNIKAL100_unmapUserKIOBuf(pageLockToken->pageLockDescriptor);
      }
      else
      {
         nNIKAL100_debugPrintf("_pageLockUserPointer: Failed to pagelock address through mapUserKIOBuf(%p, %ld), status: %ld\n", pageAlignedInPointer, size, (long)status);
      }

      
      nNIKAL100_unmapPhysicalFromKernel(*outPageLockedPointer);
      *outPageLockedPointer = NULL;
   }
   else
   {
      nNIKAL100_debugPrintf("_pageLockUserPointer: Failed to reserve kernel virtual address through ioremap(address: %p, size: %ld)\n", reusableIORemapAddress, size);
      status = nNIKAL100_statusSelect(status, nNIKAL100_kStatusMemoryFull);
   }

   up(&nNIKAL100_sPageLockOperationSema);
   return status;
}


nNIKAL100_cc nNIKAL100_tStatus nNIKAL100_pageUnlockUserPointer(nNIKAL100_tMemPageLockToken * pageLockToken)
{
   
   down(&nNIKAL100_sPageLockOperationSema);

   
   flush_cache_all();

   
   nNIKAL100_setKernelPageTable(&(pageLockToken->originalPageTableEntry),
      pageLockToken->pageLockedPointer, pageLockToken->bufferSize);
   nNIKAL100_unsetVMANoCopy(pageLockToken);
   nNIKAL100_unmapUserKIOBuf(pageLockToken->pageLockDescriptor);
   nNIKAL100_unmapPhysicalFromKernel(pageLockToken->pageLockedPointer);

   up(&nNIKAL100_sPageLockOperationSema);

   return nNIKAL100_kStatusSuccess;
}

nNIKAL100_cc const nNIKAL100_tText **nNIVersion_NIKAL_getVersionStringArray(void)
{
   return nNIVersion_versionStringArray;
}



nNIKAL100_cc nNIKAL100_tUPtr nNIKAL120_checkStackUsage(void)
{
   nNIKAL100_tUPtr stackUsage = 
      nNIKAL120_mRoundUp(nNIKAL120_sStackSize, PAGE_SIZE);
   
   stackUsage = stackUsage - ((nNIKAL100_tUPtr)(&stackUsage) & (stackUsage-1));

   if (stackUsage > nNIKAL120_sMaxStackUsageSeen)
   {
      nNIKAL120_sMaxStackUsageSeen = stackUsage;
      if (stackUsage > nNIKAL120_kMinStackUsageForWarning)
      {
         nNIKAL100_debugPrintf("nNIKAL120_checkStackUsage(): WARNING: stackUsage: %lu\n", stackUsage);
         /** Uncomment the line below to get the stack dump when stack is too deep **/
         /** dump_stack(); **/
      }
   }
   return stackUsage;
}


nNIKAL100_cc nNIKAL100_tU32 nNIKAL170_getCallStack(u32 callingLevel, u32 startLevel, void **retAddrBuffer)
{
#if defined(nNIKAL100_kDebuggingIsActive) && defined(__x86_64__)
   nNIKAL100_tU32 i;
   nNIKAL100_tU32 j;
   void ** currentAddr;
   void ** bottomOfStack;
   nLinux_driver * module;

   
   for (i=0;i<callingLevel;i++)
   {
      retAddrBuffer[i] = NULL;
   }

   currentAddr = (void **)nNIKAL120_mRoundUp((nNIKAL100_tUPtr)&callingLevel, sizeof(void*));

   bottomOfStack = (void **)nNIKAL120_mRoundUp((nNIKAL100_tUPtr)currentAddr,
                                               nNIKAL120_mRoundUp(nNIKAL120_sStackSize, PAGE_SIZE));

   for (i = j = 0; j < callingLevel && currentAddr < bottomOfStack; ++currentAddr)
   {
      
      if ((*currentAddr >= THIS_MODULE->module_init &&
           *currentAddr < THIS_MODULE->module_init + THIS_MODULE->init_text_size) ||
          (*currentAddr >= THIS_MODULE->module_core &&
           *currentAddr < THIS_MODULE->module_core + THIS_MODULE->core_text_size))
      {
         if (i >= startLevel)
            retAddrBuffer[j++]=*currentAddr;
         ++i;
         continue;
      }
      
      list_for_each_entry(module, &(THIS_MODULE->list), list)
      {
         
         if (module->mkobj.mod == module)
         {
            
            if ((*currentAddr >= module->module_init &&
                 *currentAddr < module->module_init + module->init_text_size) ||
                (*currentAddr >= module->module_core &&
                 *currentAddr < module->module_core + module->core_text_size))
            {
               if (i >= startLevel)
                  retAddrBuffer[j++]=*currentAddr;
               ++i;
               break;
            }
         }
      }
   }
   return j;
#else
   nNIKAL100_tU32 i;
   for (i=0;i<callingLevel;i++)
   {
      retAddrBuffer[i] = NULL;
   }
   return 0;
#endif
}

module_init(nNIKAL100_initDriver);
module_exit(nNIKAL100_cleanupDriver);

#ifdef EXPORT_SYMBOL_NOVERS
#define nNIKAL100_mExportSymbol EXPORT_SYMBOL_NOVERS
#else
#define nNIKAL100_mExportSymbol EXPORT_SYMBOL
#endif
nNIKAL100_mExportSymbol(nNIKAL100_initializeSingleUseEvent);
nNIKAL100_mExportSymbol(nNIKAL100_relinquishInterrupt);
nNIKAL100_mExportSymbol(nNIKAL100_vsprintf);
nNIKAL100_mExportSymbol(nNIKAL100_createDPC);
nNIKAL100_mExportSymbol(nNIKAL100_getKernelVersion);
nNIKAL100_mExportSymbol(nNIKAL100_setPCIDeviceIDTableElement);
nNIKAL100_mExportSymbol(nNIKAL100_setUSBDeviceIDTableElement);
nNIKAL100_mExportSymbol(nNIKAL100_memoryRead16);
nNIKAL100_mExportSymbol(nNIKAL100_getPhysicalAddress);
nNIKAL100_mExportSymbol(nNIKAL100_ioWrite16);
nNIKAL100_mExportSymbol(nNIKAL100_enumeratePCIBuses);
nNIKAL100_mExportSymbol(nNIKAL100_sleepTimeout);
nNIKAL100_mExportSymbol(nNIKAL100_getTimerInterval);
nNIKAL100_mExportSymbol(nNIKAL100_sleepTimeoutInterruptible);
nNIKAL100_mExportSymbol(nNIKAL100_createPCIDeviceIDTable);
nNIKAL100_mExportSymbol(nNIKAL100_createUSBDeviceIDTable);
nNIKAL100_mExportSymbol(nNIKAL100_malloc);
nNIKAL100_mExportSymbol(nNIKAL100_malloc32BitPhysicalContiguous);
nNIKAL100_mExportSymbol(nNIKAL100_registerDriver);
nNIKAL100_mExportSymbol(nNIKAL100_getTimerCount);
nNIKAL100_mExportSymbol(nNIKAL110_getTimerCount);
nNIKAL100_mExportSymbol(nNIKAL100_pciConfigRead16);
nNIKAL100_mExportSymbol(nNIKAL100_ioRead32);
nNIKAL100_mExportSymbol(nNIKAL100_destroyPCIDeviceIDTable);
nNIKAL100_mExportSymbol(nNIKAL100_destroyUSBDeviceIDTable);
nNIKAL100_mExportSymbol(nNIKAL100_ioWrite8);
nNIKAL100_mExportSymbol(nNIKAL100_unmapPhysicalFromKernel);
nNIKAL100_mExportSymbol(nNIKAL100_acquireSpinLockDPC);
nNIKAL100_mExportSymbol(nNIKAL110_acquireSpinLockInterrupt);
nNIKAL100_mExportSymbol(nNIKAL100_acquireSpinLockInterrupt);
nNIKAL100_mExportSymbol(nNIKAL100_acquireSemaphoreInterruptible);
nNIKAL100_mExportSymbol(nNIKAL100_getSingleUseEventSize);
nNIKAL100_mExportSymbol(nNIKAL100_destroyDPC);
nNIKAL100_mExportSymbol(nNIKAL100_getPhysicalMemorySize);
nNIKAL100_mExportSymbol(nNIKAL100_getPageSize);
nNIKAL100_mExportSymbol(nNIKAL100_printToDebugger);
nNIKAL100_mExportSymbol(nNIKAL100_vsnPrintf);
nNIKAL100_mExportSymbol(nNIKAL100_replicateOutboundIOControlBuffer);
nNIKAL100_mExportSymbol(nNIKAL100_pciConfigWrite32);
nNIKAL100_mExportSymbol(nNIKAL100_pciConfigWrite16);
nNIKAL100_mExportSymbol(nNIKAL100_releaseSemaphore);
nNIKAL100_mExportSymbol(nNIKAL100_acquireSemaphore);
nNIKAL100_mExportSymbol(nNIKAL100_getTimeOfDay);
nNIKAL100_mExportSymbol(nNIKAL100_memoryRead8);
nNIKAL100_mExportSymbol(nNIKAL100_releaseSpinLockDPC);
nNIKAL100_mExportSymbol(nNIKAL100_registerPCIDriver);
nNIKAL100_mExportSymbol(nNIKAL100_registerUSBDriver);
nNIKAL100_mExportSymbol(nNIKAL100_getThreadID);
nNIKAL100_mExportSymbol(nNIKAL100_pciConfigRead8);
nNIKAL100_mExportSymbol(nNIKAL100_free);
nNIKAL100_mExportSymbol(nNIKAL100_free32BitPhysicalContiguous);
nNIKAL100_mExportSymbol(nNIKAL100_waitForSingleUseEventTimeoutInterruptible);
nNIKAL100_mExportSymbol(nNIKAL100_scheduleDPC);
nNIKAL100_mExportSymbol(nNIKAL100_pciConfigRead32);
nNIKAL100_mExportSymbol(nNIKAL100_ioRead16);
nNIKAL100_mExportSymbol(nNIKAL100_pciConfigWrite8);
nNIKAL100_mExportSymbol(nNIKAL100_memoryWrite32);
nNIKAL100_mExportSymbol(nNIKAL100_getTimeOfDayInterval);
nNIKAL100_mExportSymbol(nNIKAL100_createSpinLock);
nNIKAL100_mExportSymbol(nNIKAL100_ioWrite32);
nNIKAL100_mExportSymbol(nNIKAL100_reserveInterrupt);
nNIKAL100_mExportSymbol(nNIKAL100_createSystemThread);
nNIKAL100_mExportSymbol(nNIKAL100_acquireSemaphoreZeroTimeout);
nNIKAL100_mExportSymbol(nNIKAL100_getThreadHandle);
nNIKAL100_mExportSymbol(nNIKAL100_memoryWrite16);
nNIKAL100_mExportSymbol(nNIKAL100_pageLockUserPointer);
nNIKAL100_mExportSymbol(nNIKAL100_destroySemaphore);
nNIKAL100_mExportSymbol(nNIKAL100_destroySpinLock);
nNIKAL100_mExportSymbol(nNIKAL100_unregisterDriver);
nNIKAL100_mExportSymbol(nNIKAL100_ioRead8);
nNIKAL100_mExportSymbol(nNIKAL100_memoryRead32);
nNIKAL100_mExportSymbol(nNIKAL100_createSemaphore);
nNIKAL100_mExportSymbol(nNIKAL100_getNumberOfActiveProcessors);
nNIKAL100_mExportSymbol(nNIKAL100_incrementDriverRefcount);
nNIKAL100_mExportSymbol(nNIKAL100_mallocContiguous);
nNIKAL100_mExportSymbol(nNIKAL100_pageUnlockUserPointer);
nNIKAL100_mExportSymbol(nNIKAL100_mapPhysicalToKernel);
nNIKAL100_mExportSymbol(nNIKAL100_waitForSingleUseEvent);
nNIKAL100_mExportSymbol(nNIKAL100_loadDriver);
nNIKAL100_mExportSymbol(nNIKAL100_releaseSingleUseEvent);
nNIKAL100_mExportSymbol(nNIKAL100_waitForSingleUseEventInterruptible);
nNIKAL100_mExportSymbol(nNIKAL100_unmapKernelFromUser);
nNIKAL100_mExportSymbol(nNIKAL100_replicateInboundIOControlBuffer);
nNIKAL100_mExportSymbol(nNIKAL110_releaseSpinLockInterrupt);
nNIKAL100_mExportSymbol(nNIKAL100_releaseSpinLockInterrupt);
nNIKAL100_mExportSymbol(nNIKAL100_mapKernelToUser);
nNIKAL100_mExportSymbol(nNIKAL100_unregisterPCIDriver);
nNIKAL100_mExportSymbol(nNIKAL100_unregisterUSBDriver);
nNIKAL100_mExportSymbol(nNIKAL100_decrementDriverRefcount);
nNIKAL100_mExportSymbol(nNIKAL100_waitForSingleUseEventTimeout);
nNIKAL100_mExportSymbol(nNIKAL100_memoryWrite8);
nNIKAL100_mExportSymbol(nNIKAL100_usbAllocateURB);
nNIKAL100_mExportSymbol(nNIKAL100_usbFreeURB);
nNIKAL100_mExportSymbol(nNIKAL100_usbSubmitURB);
nNIKAL100_mExportSymbol(nNIKAL100_usbUnlinkURB);
nNIKAL100_mExportSymbol(nNIKAL100_usbGetDeviceDescriptor);
nNIKAL100_mExportSymbol(nNIKAL100_usbGetConfigurationDescriptor);
nNIKAL100_mExportSymbol(nNIKAL100_usbGetInterfaceDescriptor);
nNIKAL100_mExportSymbol(nNIKAL100_usbGetEndpointDescriptor);
nNIKAL100_mExportSymbol(nNIKAL100_usbSynchronousControlRequest);
nNIKAL100_mExportSymbol(nNIKAL100_usbSetConfiguration);
nNIKAL100_mExportSymbol(nNIKAL100_usbGetConfiguration);
nNIKAL100_mExportSymbol(nNIKAL100_usbSetInterface);
nNIKAL100_mExportSymbol(nNIKAL100_usbGetInterface);
nNIKAL100_mExportSymbol(nNIVersion_NIKAL_getVersionStringArray);
nNIKAL100_mExportSymbol(nNIKAL120_checkStackUsage);
nNIKAL100_mExportSymbol(nNIKAL170_getCallStack);
nNIKAL100_mExportSymbol(nNIKAL150_createMutex);
nNIKAL100_mExportSymbol(nNIKAL150_destroyMutex);
nNIKAL100_mExportSymbol(nNIKAL150_acquireMutex);
nNIKAL100_mExportSymbol(nNIKAL150_acquireMutexInterruptible);
nNIKAL100_mExportSymbol(nNIKAL150_acquireMutexZeroTimeout);
nNIKAL100_mExportSymbol(nNIKAL150_releaseMutex);
nNIKAL100_mExportSymbol(nNIKAL180_memCpy);
nNIKAL100_mExportSymbol(nNIKAL180_memMove);
